/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */

tree grammar VerilogTreeBuilder;

options {
  tokenVocab   = VerilogParser;
  ASTLabelType = CommonTree;

  // Description of the expression tree description is left-recursive:
  // to avoid the warning, backtracking is turned on.
  backtrack = true;
}

@header {
  package ru.ispras.verilog.parser.grammar;

  import java.util.ArrayList;
  import java.util.EnumSet;
  import java.util.List;
  import java.util.Stack;

  import ru.ispras.fortress.expression.*;
  import ru.ispras.fortress.solver.function.*;
  import ru.ispras.verilog.parser.VerilogFrontend;
  import ru.ispras.verilog.parser.VerilogLogger;
  import ru.ispras.verilog.parser.model.*;
  import ru.ispras.verilog.parser.model.basis.*;
  import ru.ispras.verilog.parser.model.util.*;
}

@members {
  // The current scope.
  private final Stack<VerilogNode> scope = new Stack<>();

  // Reference to the parent Verilog front-end.
  private VerilogFrontend frontend;

  /**
   * Constructs the Verilog tree builder.
   *
   * @param nodes    the node stream to be transformed into a syntax tree.
   * @param frontend the reference to the parent Verilog front-end.
   */
  public VerilogTreeBuilder(final CommonTreeNodeStream nodes, final VerilogFrontend frontend) {
    super(nodes);
    this.frontend = frontend;
  }

  @Override
  public void emitErrorMessage(final String msg) {
    VerilogLogger.error(msg);
  }
}

startRule returns [VerilogCode model]
    : ast_root {$model = $ast_root.model;}
;

ast_root returns [VerilogCode model = new VerilogCode()]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_ROOT (ast_unit {$model.add($ast_unit.model);
                            $model.setFileName(this.frontend.getCurrentFileName());})*)
;

ast_unit returns [VerilogNode model]
    : ^(AST_ATTRIBUTED_UNIT ( ast_module {$model = $ast_module.model;})
        ast_attributes {$model.setAttributes($ast_attributes.model);}
      )
;

ast_module returns [VerilogModule model = new VerilogModule(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_MODULE
        ast_identifier
            {$model.setName($ast_identifier.model);}
        ^(AST_MODULE_TYPE ( KW_MODULE      {$model.setModule();}
                          | KW_MACROMODULE {$model.setMacromodule();}
                          | KW_PRIMITIVE   {$model.setUserPrimitive();}))
        // IEEE 1364-1995: ports are listed in the signature and then declared.
        ^(AST_INTERFACE (ast_port
            {$model.addPort($ast_port.model);})*)
        // IEEE 1364-2001: ports are declared in the signature.
        ^(AST_MODULE_ITEMS (ast_item
            {
              for (final VerilogNode item : $ast_item.model) {
                // Declared ports are added into the signature.
                // Declarations of the same variable are merged.
                if (item.isDeclaration())
                    { $model.addDeclaration((VerilogDeclaration) item); }
                else
                    { $model.add(item); }
              }
            })*)
    )
;

ast_port returns [VerilogPort model = new VerilogPort(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_PORT
        ast_identifier
            {$model.setName($ast_identifier.model);}
        ^(AST_REFERENCES (ast_reference
            {$model.addReference($ast_reference.model);})*)
    )
;

ast_item returns [List<VerilogNode> model = new ArrayList<>()]
    : ^(AST_ATTRIBUTED_ITEM ( ast_declaration      {$model.add($ast_declaration.model);}
                            | ast_generate         {$model.add($ast_generate.model);}
                            | ast_loop_generate    {$model.add($ast_loop_generate.model);}
                            | ast_if_generate      {$model.add($ast_if_generate.model);}
                            | ast_case_generate    {$model.add($ast_case_generate.model);}
                            | ast_assign           {$model.add($ast_assign.model);}
                            | ast_process          {$model.add($ast_process.model);}
                            | ast_instantiation    {$model.add($ast_instantiation.model);}
                            | ast_procedure        {$model.add($ast_procedure.model);}
                            | ast_table            {$model.add($ast_table.model);}
                            | ast_specify          {$model.add($ast_specify.model);}
                            | ast_define_parameter {$model.add($ast_define_parameter.model);})+
        ast_attributes {
          for (final VerilogNode item : model) {
            item.setAttributes($ast_attributes.model);
          }
        }
      )
;

ast_declaration returns [VerilogDeclaration model = new VerilogDeclaration(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_DECLARATION
        ^(AST_ELEMENT
            ast_identifier
                {$model.setName($ast_identifier.model);}
            ^(AST_DIMENSIONS (ast_range
                {$model.addDimension($ast_range.model);})*)
            ^(AST_EXPRESSION (ast_extended_expression
                {$model.setExpression($ast_extended_expression.model);})?))
        ^(AST_ELEMENT_TYPE ( KW_INPUT      {$model.setInput();}
                           | KW_OUTPUT     {$model.setOutput();}
                           | KW_INOUT      {$model.setInout();}
                           | KW_EVENT      {$model.setEvent();}
                           | KW_PARAMETER  {$model.setParameter();}
                           | KW_SPECPARAM  {$model.setSpecparam();}
                           | KW_LOCALPARAM {$model.setLocalparam();}
                           | KW_GENVAR     {$model.setGenvar();})?)
        ast_element_type
            {$model.setElementType($ast_element_type.model);}
    )
;

ast_element_type returns [VerilogElementType model = new VerilogElementType()]
    : ^(AST_TYPE
        ^(AST_DATA_TYPE ( KW_INTEGER  {$model.setInteger();}
                        | KW_REAL     {$model.setReal();}
                        | KW_TIME     {$model.setTime();}
                        | KW_REALTIME {$model.setRealtime();}
                        | KW_SUPPLY0  {$model.setSupply0();}
                        | KW_SUPPLY1  {$model.setSupply1();}
                        | KW_TRI      {$model.setTri();}
                        | KW_TRIAND   {$model.setTriand();}
                        | KW_TRIOR    {$model.setTrior();}
                        | KW_TRI0     {$model.setTri0();}
                        | KW_TRI1     {$model.setTri1();}
                        | KW_UWIRE    {$model.setUwire();}
                        | KW_WIRE     {$model.setWire();}
                        | KW_WAND     {$model.setWand();}
                        | KW_WOR      {$model.setWor();}
                        | KW_TRIREG   {$model.setTrireg();}
                        | KW_REG      {$model.setReg();})?)
        ast_strength                  {$model.setStrength($ast_strength.model);}
        ^(AST_MODIFIERS ( KW_VECTORED {$model.setVectored();}
                        | KW_SCALARED {$model.setScalared();}
                        | KW_SIGNED   {$model.setSigned();})*)
        ^(AST_DIMENSIONS (ast_range   {$model.setRange($ast_range.model);})?)
      )
;

ast_generate returns [VerilogGenerate model = new VerilogGenerate(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_GENERATE (ast_item {$model.addAll($ast_item.model);})*)
;

ast_loop_generate returns [VerilogLoopGenerate model = new VerilogLoopGenerate(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_LOOP_GENERATE
        init=ast_assignment
            {$model.setInitialization($init.model);}
        ^(AST_EXPRESSION ast_expression
            {$model.setExpression($ast_expression.model);})
        step=ast_assignment
            {$model.setIteration($step.model);}
        ^(AST_GENERATE_ITEM ast_generate_block
            {$model.setGenerate($ast_generate_block.model);})
      )
;

ast_if_generate returns [VerilogIfGenerate model = new VerilogIfGenerate(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_IF_GENERATE
        ^(AST_EXPRESSION ast_expression
            {$model.setExpression($ast_expression.model);})
        ^(AST_GENERATE_ITEM then_generate=ast_generate_block
            {$model.setThenGenerate($then_generate.model);})
        ^(AST_GENERATE_ITEM (else_generate=ast_generate_block
            {$model.setElseGenerate($else_generate.model);})?)
      )
;

ast_case_generate returns [VerilogCaseGenerate model = new VerilogCaseGenerate(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_CASE_GENERATE
        ^(AST_EXPRESSION ast_expression
            {$model.setExpression($ast_expression.model);})
        ^(AST_CASE_GENERATE_ITEMS (ast_case_generate_item
            {$model.addCase($ast_case_generate_item.model);})+)
      )
;

ast_case_generate_item returns [VerilogCaseGenerateItem model = new VerilogCaseGenerateItem(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_CASE_GENERATE_ITEM
        ^(AST_EXPRESSIONS (ast_expression
            {$model.addExpression($ast_expression.model);})*)
        ^(AST_GENERATE_ITEM ast_generate_block
            {$model.setGenerate($ast_generate_block.model);})
      )
;

ast_generate_block returns [VerilogBlockGenerate model = new VerilogBlockGenerate(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_BLOCK_GENERATE
        ast_identifier
            {$model.setName($ast_identifier.model);}
        ^(AST_GENERATE_ITEMS (ast_item
            {$model.addAll($ast_item.model);})*)
      )
;

ast_specify returns [VerilogSpecify model = new VerilogSpecify(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_SPECIFY
          (ast_specify_item {$model.add($ast_specify_item.model);})*
      )
;

ast_specify_item returns [VerilogNode model]
    : ast_declaration      {$model = $ast_declaration.model;}
    | ast_pulsestyle       {$model = $ast_pulsestyle.model;}
    | ast_showcancelled    {$model = $ast_showcancelled.model;}
    | ast_path_declaration {$model = $ast_path_declaration.model;}
    | ast_task_statement   {$model = $ast_task_statement.model;}
;

ast_pulsestyle returns [VerilogPulseStyle model = new VerilogPulseStyle(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_PULSESTYLE
          ^(AST_PULSESTYLE_TYPE ( KW_PULSESTYLE_ONEVENT  {$model.setOnEvent();}
                                | KW_PULSESTYLE_ONDETECT {$model.setOnDetect();}))
          ^(AST_REFERENCES (ast_reference
              {$model.addReference($ast_reference.model);})+)
      )
;

ast_showcancelled returns [VerilogShowCancelled model = new VerilogShowCancelled(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_SHOWCANCELLED
          ^(AST_SHOWCANCELLED_TYPE ( KW_SHOWCANCELLED   {$model.setShow();}
                                   | KW_NOSHOWCANCELLED {$model.setNoShow();}))
          ^(AST_REFERENCES (ast_reference
              {$model.addReference($ast_reference.model);})+)
      )
;

ast_path_declaration returns [VerilogPathDeclaration model = new VerilogPathDeclaration(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_PATH_DECLARATION
          ^(AST_PATH_IF ( KW_IF     {$model.setIf();}
                        | KW_IFNONE {$model.setIfNone();})?)
          ^(AST_EXPRESSION (ast_expression
              {$model.setExpression($ast_expression.model);})?)
          ^(AST_PATH_REFERENCE (ast_path_description
              {$model.setDescription($ast_path_description.model);})?)
          ast_delay
              {$model.setDelay($ast_delay.model);}
          ^(AST_PATH_THEN (then=ast_path_declaration
              {$model.setThen($then.model);})?)
      )
;

ast_path_description returns [VerilogPathDescription model = new VerilogPathDescription()]
    : ^(AST_PATH_DESCRIPTION
          ^(AST_PATH_TYPE ( PPATH {$model.setParallel();}
                          | FPATH {$model.setFull();}))
          ast_edge
              {$model.setEdge($ast_edge.model);}
          ^(AST_REFERENCES (in=ast_reference
            {$model.addInput($in.model);})+)
          ^(AST_OPERATION ( PLUS  {$model.setPlus();}
                          | MINUS {$model.setMinus();})?)
          ^(AST_REFERENCES (out=ast_reference
            {$model.addOutput($out.model);})+)
          ^(AST_EXPRESSION (ast_expression
              {$model.setExpression($ast_expression.model);})?)
      )
;

ast_assign returns [VerilogAssign model = new VerilogAssign(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_CONTINUOUS_ASSIGNMENT
        ast_strength
            {$model.setStrength($ast_strength.model);}
        ast_delay
            {$model.setDelay($ast_delay.model);}
        ast_assignment
            {$model.setAssignment($ast_assignment.model);}
      )
;

ast_strength returns [VerilogStrength model = new VerilogStrength()]
    : ^(AST_STRENGTH ( KW_SUPPLY0 {model.setSupply0();}
                     | KW_SUPPLY1 {model.setSupply1();}
                     | KW_STRONG0 {model.setStrong0();}
                     | KW_STRONG1 {model.setStrong1();}
                     | KW_PULL0   {model.setPull0();}
                     | KW_PULL1   {model.setPull1();}
                     | KW_LARGE   {model.setLarge();}
                     | KW_LARGE0  {model.setLarge0();}
                     | KW_LARGE1  {model.setLarge1();}
                     | KW_WEAK0   {model.setWeak0();}
                     | KW_WEAK1   {model.setWeak1();}
                     | KW_MEDIUM  {model.setMedium();}
                     | KW_MEDIUM0 {model.setMedium0();}
                     | KW_MEDIUM1 {model.setMedium1();}
                     | KW_SMALL   {model.setSmall();}
                     | KW_SMALL0  {model.setSmall0();}
                     | KW_SMALL1  {model.setSmall1();}
                     | KW_HIGHZ0  {model.setHighZ0();}
                     | KW_HIGHZ1  {model.setHighZ1();})*
      )
;

ast_define_parameter returns [VerilogDefineParameter model = new VerilogDefineParameter(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_DEFINE_PARAMETER
        ast_path
            {$model.setPath($ast_path.model);}
        ^(AST_EXPRESSION (ast_extended_expression
            {$model.setExpression($ast_extended_expression.model);}))
      )
;

ast_process returns [VerilogActivity model = new VerilogActivity(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_ACTIVITY
        ^(AST_ACTIVITY_TYPE ( KW_INITIAL {$model.setInitial();}
                            | KW_ALWAYS  {$model.setAlways();}))
        ^(AST_STATEMENT ast_statement
            {$model.setStatement($ast_statement.model);})
      )
;

ast_instantiation returns [VerilogInstantiation model = new VerilogInstantiation(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_INSTANTIATION
        module=ast_identifier
            {$model.setModuleName($module.model);}
        ast_strength
            {$model.setStrength($ast_strength.model);}
        ast_delay
            {$model.setDelay($ast_delay.model);}
        ^(AST_PARAMETRIZATION (ast_assignment
            {$model.addParameter($ast_assignment.model);})*)
        ^(AST_INSTANCE
            instance=ast_identifier
                {$model.setName($instance.model);}
            ^(AST_DIMENSIONS (ast_range
                {$model.setRange($ast_range.model);})?)
            // NOTE: Empty list of port connections is enabled, but it is represented as (<empty>).
            ^(AST_PORT_CONNECTIONS (ast_port_connection
               {$model.addConnection($ast_port_connection.model);})+))
      )
;

ast_port_connection returns [VerilogPortConnection model = new VerilogPortConnection(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_PORT_CONNECTION
        ast_identifier
            {$model.setName($ast_identifier.model);}
        ^(AST_EXPRESSION (ast_expression
            {$model.setExpression($ast_expression.model);})?)
        ast_attributes
            {$model.setAttributes($ast_attributes.model);}
      )
;

ast_procedure returns [VerilogProcedure model = new VerilogProcedure(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_PROCEDURE
        ast_identifier
            {$model.setName($ast_identifier.model);}
        ^(AST_PROCEDURE_TYPE ( KW_TASK     {$model.setTask();}
                             | KW_FUNCTION {$model.setFunction();}))
        ast_element_type
            {
              if ($model.isFunction()) {
                $model.setOutput($ast_identifier.model, $ast_element_type.model);
              }
            }
        ^(AST_AUTOMATIC (KW_AUTOMATIC
            {$model.setAutomatic();})?)
        ^(AST_PROCEDURE_ITEMS (ast_item
            {
              for (final VerilogNode item : $ast_item.model) {
                $model.add(item);
                if (item.isDeclaration())
                    { $model.addPort((VerilogDeclaration) item); }
              }
            })*)
        ^(AST_STATEMENT ast_statement
            {$model.setStatement($ast_statement.model);})
      )
;

ast_table returns [VerilogTable model = new VerilogTable(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_TABLE
          (ast_table_entry
              {$model.add($ast_table_entry.model);})+
      )
;

ast_table_entry returns [VerilogTableEntry model = new VerilogTableEntry(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_TABLE_ENTRY
          ^(AST_TABLE_LEVELS
            (
              input1=ast_level_symbol
              {
                for (int i = 0; i < $input1.model.length(); i++)
                  {$model.addInput1($input1.model.substring(i, i + 1));}
              }
            )*
          )
          ^(AST_TABLE_EDGE
            (
              ast_edge_symbol
                {$model.addEdge($ast_edge_symbol.model);}
            )*
          )
          ^(AST_TABLE_LEVELS
            (
              input2=ast_level_symbol
              {
                for (int i = 0; i < $input2.model.length(); i++) {
                  $model.addInput2($input2.model.substring(i, i + 1));
                }
              }
            )*
          )
          ^(AST_TABLE_LEVELS
            (
              output=ast_level_symbol
              {
                for (int i = 0; i < $output.model.length(); i++)
                  {$model.addOutput($output.model.substring(i, i + 1));}
              }
            )+
          )
      )
;

ast_level_symbol returns [String model]
    : name=(IDENT | NUMBER | QMARK | MINUS) {$model=$name.text;}
;

ast_edge_symbol returns [String model]
    : val1=ast_level_symbol
      val2=ast_level_symbol {$model=$val1.model + $val2.model;}
    | name=(IDENT | NUMBER | STAR) {$model=$name.text;}
;

ast_statement returns [VerilogStatement model]
    : ^(AST_ATTRIBUTED_STATEMENT ( ast_null_statement    {$model = $ast_null_statement.model;}
                                 | ast_assign_statement  {$model = $ast_assign_statement.model;}
                                 | ast_task_statement    {$model = $ast_task_statement.model;}
                                 | ast_wait_statement    {$model = $ast_wait_statement.model;}
                                 | ast_disable_statement {$model = $ast_disable_statement.model;}
                                 | ast_trigger_statement {$model = $ast_trigger_statement.model;}
                                 | ast_delayed_statement {$model = $ast_delayed_statement.model;}
                                 | ast_if_statement      {$model = $ast_if_statement.model;}
                                 | ast_case_statement    {$model = $ast_case_statement.model;}
                                 | ast_loop_statement    {$model = $ast_loop_statement.model;}
                                 | ast_block_statement   {$model = $ast_block_statement.model;})
        ast_attributes {$model.setAttributes($ast_attributes.model);}
      )
;

ast_null_statement returns [VerilogNullStatement model = new VerilogNullStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : AST_NULL_STATEMENT
;

ast_assign_statement returns [VerilogAssignStatement model = new VerilogAssignStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_ASSIGN_STATEMENT
        ^(AST_ASSIGN_TYPE ( ASSIGN      {$model.setBlocking();}
                          | LESSEQ      {$model.setNonBlocking();}
                          | KW_ASSIGN   {$model.setAssign();}
                          | KW_FORCE    {$model.setForce();}
                          | KW_DEASSIGN {$model.setDeassign();}
                          | KW_RELEASE  {$model.setRelease();}))
        ast_event_control
            {$model.setEventControl($ast_event_control.model);}
        ast_assignment
            {$model.setAssignment($ast_assignment.model);}
      )
;

ast_task_statement returns [VerilogTaskStatement model = new VerilogTaskStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_TASK_STATEMENT
        ast_path
            {$model.setPath($ast_path.model);}
        ^(AST_EXPRESSIONS (ast_expression
            {$model.addArgument($ast_expression.model);})*)
      )
;

ast_wait_statement returns [VerilogWaitStatement model = new VerilogWaitStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_WAIT_STATEMENT
        ^(AST_EXPRESSION ast_expression
            {$model.setExpression($ast_expression.model);})
        ^(AST_STATEMENT  ast_statement
            {$model.setStatement($ast_statement.model);})
      )
;

ast_disable_statement returns [VerilogDisableStatement model = new VerilogDisableStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_DISABLE_STATEMENT
        ast_path
            {$model.setPath($ast_path.model);}
      )
;

ast_trigger_statement returns [VerilogTriggerStatement model = new VerilogTriggerStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_TRIGGER_STATEMENT
        ast_path
            {$model.setPath($ast_path.model);}
        ^(AST_EXPRESSIONS (ast_expression
            {$model.addArgument($ast_expression.model);})*)
      )
;

ast_delayed_statement returns [VerilogDelayedStatement model = new VerilogDelayedStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_DELAYED_STATEMENT
        ast_event_control
            {$model.setEventControl($ast_event_control.model);}
        ^(AST_STATEMENT ast_statement
            {$model.setStatement($ast_statement.model);})
      )
;

ast_if_statement returns [VerilogIfStatement model = new VerilogIfStatement(scope.peek())]
@init  { scope.push(model); boolean hasElseBranch = false;}
@after { if (!hasElseBranch)
           {$model.setElseStatement(new VerilogBlockStatement(scope.peek()));}
         scope.pop();
       }
    : ^(AST_IF_STATEMENT
        ^(AST_EXPRESSION ast_expression
            {$model.setExpression($ast_expression.model);})
        ^(AST_STATEMENT then_statement=ast_statement
            {$model.setThenStatement($then_statement.model);})
        ^(AST_STATEMENT (else_statement=ast_statement
            {$model.setElseStatement($else_statement.model); hasElseBranch = true;})?)
      )
;

ast_case_statement returns [VerilogCaseStatement model = new VerilogCaseStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_CASE_STATEMENT
        ^(AST_CASE_TYPE ( KW_CASE  {$model.setCase();}
                        | KW_CASEX {$model.setCaseX();}
                        | KW_CASEZ {$model.setCaseZ();}))
        ^(AST_EXPRESSION expression=ast_expression
            {$model.setExpression($expression.model);})
        ^(AST_CASE_STATEMENT_ITEMS (ast_case_statement_item
	    {$model.addCase($ast_case_statement_item.model);})+)
      )
;

ast_loop_statement returns [VerilogLoopStatement model = new VerilogLoopStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_LOOP_STATEMENT
        ^(AST_LOOP_TYPE ( KW_FOREVER {$model.setForever();}
                        | KW_REPEAT  {$model.setRepeat();}
                        | KW_WHILE   {$model.setWhile();}
                        | KW_FOR     {$model.setFor();}))
        (init=ast_assignment
            {$model.setInitialization($init.model);})?
        ^(AST_EXPRESSION (ast_expression
            {$model.setExpression($ast_expression.model);})?)
        (step=ast_assignment
            {$model.setIteration($step.model);})?
        ^(AST_STATEMENT  ast_statement
            {$model.setStatement($ast_statement.model);})
      )
;

ast_block_statement returns [VerilogBlockStatement model = new VerilogBlockStatement(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_BLOCK_STATEMENT
        ast_identifier
            {$model.setName($ast_identifier.model);}
        ^(AST_BLOCK_TYPE  ( KW_BEGIN {$model.setBegin();}
                          | KW_FORK  {$model.setFork();}))
        ^(AST_BLOCK_ITEMS (ast_item
            {$model.addAll($ast_item.model);})*)
        ^(AST_STATEMENTS  (ast_statement
            {$model.add($ast_statement.model);})*)
      )
;

ast_assignment returns [VerilogAssignment model = new VerilogAssignment(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_ASSIGNMENT
        ^(AST_REFERENCES (ast_reference
            {$model.addReference($ast_reference.model);})*)
        ^(AST_EXPRESSION (ast_extended_expression
            {$model.setRhsExpression($ast_extended_expression.model);})?)
      )
;

ast_extended_expression returns [VerilogMinTypMax model = new VerilogMinTypMax()]
    : ast_expression
        {$model.setMinExpression($ast_expression.model);}
    | ast_mintypmax_expression
        {$model.setMinTypMaxExpression($ast_mintypmax_expression.model);}
;

ast_mintypmax_expression returns [VerilogMinTypMax model = new VerilogMinTypMax()]
    : ^(AST_MINTYPMAX_EXPRESSION
        ^(AST_EXPRESSION min=ast_expression
            {$model.setMinExpression($min.model);})
        ^(AST_EXPRESSION (typ=ast_expression
            {$model.setTypExpression($typ.model);})?)
        ^(AST_EXPRESSION (max=ast_expression
            {$model.setMaxExpression($max.model);})?)
      )
;

ast_case_statement_item returns [VerilogCaseStatementItem model = new VerilogCaseStatementItem(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_CASE_STATEMENT_ITEM
        ^(AST_EXPRESSIONS (ast_expression
            {$model.addExpression($ast_expression.model);})*)
        ^(AST_STATEMENT ast_statement
            {$model.setStatement($ast_statement.model);})
      )
;

ast_range returns [VerilogRange model = new VerilogRange()]
    : ^(AST_RANGE
        ^(AST_RANGE_TYPE ( COLON {$model.setRange();}
                         | PLUS  {$model.setPlus();}
                         | MINUS {$model.setMinus();})?)
        ^(AST_EXPRESSION lhs=ast_expression
            {$model.setLeftExpression($lhs.model);})
        ^(AST_EXPRESSION (rhs=ast_expression
            {$model.setRightExpression($rhs.model);})?)
      )
;

ast_event_control returns [VerilogEventControl model = new VerilogEventControl()]
    : ^(AST_EVENT_CONTROL
        ( ast_delay         {$model.setDelay($ast_delay.model);}
        | ast_events        {$model.setEvents($ast_events.model);}
        | ast_repeat_events {$model.setRepeatEvents($ast_repeat_events.model);}
        )?
      )
;

ast_delay returns [VerilogDelay model = new VerilogDelay()]
    : ^(AST_DELAY (ast_extended_expression
        {$model.addDelay($ast_extended_expression.model);})*)
;

ast_events returns [List<VerilogEvent> model = new ArrayList<VerilogEvent>()]
    : ^(AST_EVENTS (ast_event {$model.add($ast_event.model);})*)
;

ast_event returns [VerilogEvent model = new VerilogEvent()]
    : ^(AST_EVENT
        ast_edge
            {$model.setEdge($ast_edge.model);}
        ^(AST_EXPRESSION ast_expression
            {$model.setExpression($ast_expression.model);})
      )
;

ast_edge returns [VerilogEdge model = VerilogEdge.UNDEFINED]
    : ^(AST_EDGE ( KW_POSEDGE {$model = VerilogEdge.POSITIVE;}
                 | KW_NEGEDGE {$model = VerilogEdge.NEGATIVE;})?
      )
;

ast_repeat_events returns [VerilogRepeatEvents model = new VerilogRepeatEvents()]
    : ^(AST_REPEAT_EVENTS
        ^(AST_EXPRESSION ast_expression
            {$model.setExpression($ast_expression.model);})
        ast_events
            {$model.setEvents($ast_events.model);}
      )
;

ast_expression returns [VerilogExpression model]
    : ast_parenthesis_expression {$model = $ast_parenthesis_expression.model;}
    | ast_operation              {$model = $ast_operation.model;}
;

ast_parenthesis_expression returns [VerilogExpression model]
    : LPAREN ( expr=ast_parenthesis_expression {$model = $expr.model;}
    | ast_operation                            {$model = $ast_operation.model;}
    ) RPAREN
;

ast_operation returns [VerilogExpression model]
    : ast_primary           {$model = $ast_primary.model;}
    | ast_unary_operation   {$model = $ast_unary_operation.model;}
    | ast_binary_operation  {$model = $ast_binary_operation.model;}
    | ast_ternary_operation {$model = $ast_ternary_operation.model;}
;

ast_primary returns [VerilogExpression model = new VerilogExpression()]
    : ast_number               {$model.setNode(ModelUtils.getValue($ast_number.model));}
    | ast_string               {$model.setNode(ModelUtils.getValue($ast_string.model));}
    | ast_concatenation        {$model = $ast_concatenation.model;}
    | ast_replication          {$model = $ast_replication.model;}
    | ast_function_call        {$model = $ast_function_call.model;}
    | ast_reference
      {
        final String name = $ast_reference.model.getPath().toString();
        final VerilogDeclaration entry =
          (VerilogDeclaration) scope.peek().findRecursively(name, VerilogNode.Tag.DECLARATION);
        if (entry != null) {
          $ast_reference.model.setDeclaration(entry);
        }
        $model.setReference($ast_reference.model);
      }
    | ast_mintypmax_expression {$model = $ast_mintypmax_expression.model.getMinExpression();}
;

ast_unary_operation returns [VerilogExpression model = new VerilogExpression()]
@init { Enum<?> op = null; }
    : ^(AST_UNARY_OPERATION
        ^(AST_OPERATION ( MINUS {op = StandardOperation.BVNEG;}
                        | PLUS  {op = StandardOperation.PLUS;}
                        | LNOT  {op = StandardOperation.NOT;}
                        | BNOT  {op = StandardOperation.BVNOT;}
                        | BAND  {op = StandardOperation.BVANDR;}
                        | BNAND {op = StandardOperation.BVNANDR;}
                        | BOR   {op = StandardOperation.BVORR;}
                        | BNOR  {op = StandardOperation.BVNORR;}
                        | BXOR  {op = StandardOperation.BVXORR;}
                        | BXNOR {op = StandardOperation.BVXNORR;}))
        ast_expression
            {$model.setNode(new NodeOperation(op, $ast_expression.model.getNode()));}
        ast_attributes
            {$model.setAttributes($ast_attributes.model);}
      )
;

ast_binary_operation returns [VerilogExpression model = new VerilogExpression()]
@init { Enum<?> op = null; }
    : ^(AST_BINARY_OPERATION
        ^(AST_OPERATION ( POWER     {op = StandardOperation.POWER;}
                        | STAR      {op = StandardOperation.BVMUL;}
                        | DIV       {op = StandardOperation.BVSDIV;}
                        | MOD       {op = StandardOperation.BVSMOD;}
                        | PLUS      {op = StandardOperation.BVADD;}
                        | MINUS     {op = StandardOperation.BVSUB;}
                        | LSR       {op = StandardOperation.BVLSHR;}
                        | LSL       {op = StandardOperation.BVLSHL;}
                        | ASR       {op = StandardOperation.BVASHR;}
                        | ASL       {op = StandardOperation.BVASHL;}
                        | LESS      {op = StandardOperation.BVULT;}
                        | LESSEQ    {op = StandardOperation.BVULE;}
                        | GREATER   {op = StandardOperation.BVUGT;}
                        | GREATEREQ {op = StandardOperation.BVUGE;}
                        | EQ        {op = StandardOperation.EQ;}
                        | NOTEQ     {op = StandardOperation.NOTEQ;}
                        | EQCASE    {op = StandardOperation.EQCASE;}
                        | NOTEQCASE {op = StandardOperation.NOTEQCASE;}
                        | BAND      {op = StandardOperation.BVAND;}
                        | BNAND     {op = StandardOperation.BVNAND;}
                        | BXOR      {op = StandardOperation.BVXOR;}
                        | BXNOR     {op = StandardOperation.BVXNOR;}
                        | BOR       {op = StandardOperation.BVOR;}
                        | BNOR      {op = StandardOperation.BVNOR;}
                        | LAND      {op = StandardOperation.AND;}
                        | LOR       {op = StandardOperation.OR;}))
        lhs=ast_expression
        rhs=ast_expression
            {$model.setNode(new NodeOperation(op, $lhs.model.getNode(),
                                                  $rhs.model.getNode()));}
        ast_attributes
            {$model.setAttributes($ast_attributes.model);}
      )
;

ast_ternary_operation returns [VerilogExpression model = new VerilogExpression()]
@init { Enum<?> op = null; }
    : ^(AST_TERNARY_OPERATION
        ^(AST_OPERATION QMARK {op = StandardOperation.ITE;})
        cnd=ast_expression
        lhs=ast_expression
        rhs=ast_expression
            {$model.setNode(new NodeOperation(op, $cnd.model.getNode(),
                                                  $lhs.model.getNode(),
                                                  $rhs.model.getNode()));}
        ast_attributes
            {$model.setAttributes($ast_attributes.model);}
      )
;

ast_reference returns [VerilogReference model = new VerilogReference()]
    : ^(AST_REFERENCE
        ast_path
            {$model.setPath($ast_path.model);}
        ^(AST_SELECTION (ast_range
            {$model.addIndexOrBitsSelection($ast_range.model);})*)
      )
;

ast_function_call returns [VerilogExpression model = new VerilogExpression();]
@init { List<Node> args = new ArrayList<Node>();}
    : ^(AST_FUNCTION_CALL
        ast_path
            {
              $model.setPath($ast_path.model);
              // Pseudo argument that stores are a reference to the function declaration.
              args.add(NodeValue.newUnknown("<" + $ast_path.model + ">"));
            }
        ^(AST_EXPRESSIONS (ast_expression
            {args.add($ast_expression.model.getNode());})*)
        ast_attributes
            {$model.setAttributes($ast_attributes.model);}
        {$model.setNode(new NodeOperation(StandardOperation.FUNCTION, args.toArray(new Node[]{})));}
      )
;

ast_concatenation returns [VerilogExpression model]
@init { Enum<?> op = StandardOperation.BVCONCAT; }
    : ^(AST_CONCATENATION
           head=ast_expression
              {$model = $head.model;}
          (tail=ast_expression
              {$model = new VerilogExpression(
                          new NodeOperation(op, $model.getNode(), $tail.model.getNode()));})*
      )
;

ast_replication returns [VerilogExpression model = new VerilogExpression()]
@init { Enum<?> op = StandardOperation.BVREPEAT; }
    : ^(AST_REPLICATION
        ^(AST_EXPRESSION lhs=ast_expression)
        rhs=ast_concatenation
            {$model.setNode(new NodeOperation(op, $lhs.model.getNode(),
                                                  $rhs.model.getNode()));}
      )
;

ast_path returns [VerilogPath model = new VerilogPath()]
    : ^(AST_PATH (ast_path_item
        {$model.addItem($ast_path_item.model);})+
      )
;

ast_path_item returns [VerilogPathItem model = new VerilogPathItem()]
    : ^(AST_PATH_ITEM
        ast_identifier
            {$model.setName($ast_identifier.model);}
        ^(AST_EXPRESSION (ast_expression
            {$model.setExpression($ast_expression.model);})?)
      )
;

ast_attribute returns [VerilogAttribute model = new VerilogAttribute(scope.peek())]
@init  { scope.push(model); }
@after { scope.pop(); }
    : ^(AST_ATTRIBUTE
        ast_identifier
            {$model.setName($ast_identifier.model);}
        ^(AST_EXPRESSION (ast_expression
            {$model.setExpression($ast_expression.model);})?)
      )
;

ast_attributes returns [List<VerilogAttribute> model = new ArrayList<VerilogAttribute>()]
    : ^(AST_ATTRIBUTES (ast_attribute {$model.add($ast_attribute.model);})*)
;

ast_number returns [VerilogLiteral model]
    : ^(AST_NUMBER number=NUMBER {$model = VerilogLiteral.parseNumber($number.text);})
;

ast_string returns [VerilogLiteral model]
    : ^(AST_STRING string=STRING {$model = VerilogLiteral.parseString($string.text, true);})
;

ast_identifier returns [String model]
    : ^(AST_NAME (ast_name {$model=$ast_name.model;})?)
;

ast_name returns [String model]
    : name=( IDENT
           | ESCAPED_IDENT
           | SYSTEM_IDENT
           | PATHPULSE_PARAM
           | KW_AND
           | KW_NAND
           | KW_NOR
           | KW_OR
           | KW_XOR
           | KW_XNOR
           | KW_BUF
           | KW_NOT
           | KW_BUFIF1
           | KW_BUFIF0
           | KW_NOTIF1
           | KW_NOTIF0
           | KW_CMOS
           | KW_NMOS
           | KW_PMOS
           | KW_RCMOS
           | KW_RNMOS
           | KW_RPMOS
           | KW_TRAN
           | KW_RTRAN
           | KW_TRANIF1
           | KW_RTRANIF1
           | KW_TRANIF0
           | KW_RTRANIF0
           | KW_PULLUP
           | KW_PULLDOWN ) {$model=$name.text;}
;
