/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.backends.design.typecast;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.verilog.parser.model.VerilogCaseStatement;
import ru.ispras.verilog.parser.model.VerilogCaseStatementItem;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.walker.VerilogEmptyVisitor;

import java.util.LinkedList;
import java.util.List;

/**
 * {@link VerilogTypeCastVisitor} is a key component for
 * Verilog type casting back-end.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VerilogTypeCastVisitor extends VerilogEmptyVisitor {

  /**
   * Performs type casting on {@link VerilogCaseStatement} statement.
   * <p>[IEEE Std 1364-2005] 9.5 Case statement
   * The length of all the case item expressions, as well as the case expression in the parentheses,
   * shall be made equal to the length of the longest case expression and case item expression.
   * If any of these expressions is unsigned, then all of them shall be treated as unsigned. If all
   * of these expressions are signed, then they shall be treated as signed.</p>
   * @param node The statement to be transformed.
   * @return The transformed statement.
   */
  @Override
  public Result onCaseStatementEnd(final VerilogCaseStatement node) {

    final Node expr = node.getExpression().getNode();

    // Calculate maximal size of all expression types.
    int maxSize = expr.getDataType().getSize();

    for (final VerilogCaseStatementItem item : node.getCases()) {

      for (final VerilogExpression itemExpr : item.getExpressions()) {
        final Node itemExprNode = itemExpr.getNode();
        final int itemSize = itemExprNode.getDataType().getSize();

        maxSize = Math.max(itemSize, maxSize);
      }
    }

    // Cast all the expressions to bit vector of maximal size.

    final Node newExpr = TypeConversion.coerce(expr, DataType.bitVector(maxSize));
    node.setExpression(new VerilogExpression(newExpr));

    for (final VerilogCaseStatementItem item : node.getCases()) {

      final List<VerilogExpression> expressions = new LinkedList<>();

      for (final VerilogExpression itemExpr : item.getExpressions()) {

        final Node itemExprNode = itemExpr.getNode();
        final Node newItemNode = TypeConversion.coerce(itemExprNode, DataType.bitVector(maxSize));
        expressions.add(new VerilogExpression(newItemNode));
      }

      item.removeExpressions();
      item.addExpressions(expressions);
    }

    return Result.OK;
  }
}
