/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.backends.syntax.checker;

import ru.ispras.fortress.expression.ExprTreeVisitorDefault;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.VerilogTranslator;
import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.model.VerilogAssign;
import ru.ispras.verilog.parser.model.VerilogAssignStatement;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.VerilogBlockStatement;
import ru.ispras.verilog.parser.model.VerilogCaseGenerate;
import ru.ispras.verilog.parser.model.VerilogCaseStatement;
import ru.ispras.verilog.parser.model.VerilogCode;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogDelayedStatement;
import ru.ispras.verilog.parser.model.VerilogIfGenerate;
import ru.ispras.verilog.parser.model.VerilogIfStatement;
import ru.ispras.verilog.parser.model.VerilogInstantiation;
import ru.ispras.verilog.parser.model.VerilogLoopStatement;
import ru.ispras.verilog.parser.model.VerilogModule;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogNode.Tag;
import ru.ispras.verilog.parser.model.VerilogPort;
import ru.ispras.verilog.parser.model.VerilogPortConnection;
import ru.ispras.verilog.parser.model.VerilogProcedure;
import ru.ispras.verilog.parser.model.VerilogTaskStatement;
import ru.ispras.verilog.parser.model.VerilogWaitStatement;
import ru.ispras.verilog.parser.model.basis.VerilogDelay;
import ru.ispras.verilog.parser.model.basis.VerilogElementType;
import ru.ispras.verilog.parser.model.basis.VerilogEvent;
import ru.ispras.verilog.parser.model.basis.VerilogEventControl;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.model.basis.VerilogPath;
import ru.ispras.verilog.parser.model.basis.VerilogPathItem;
import ru.ispras.verilog.parser.model.basis.VerilogRange;
import ru.ispras.verilog.parser.model.basis.VerilogReference;
import ru.ispras.verilog.parser.model.util.ModelUtils;
import ru.ispras.verilog.parser.transformer.VerilogTransformerVariableRename;
import ru.ispras.verilog.parser.walker.VerilogEmptyVisitor;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * {@link VerilogStaticChecker} implements a Verilog static checker.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogStaticChecker extends VerilogEmptyVisitor {

  /**
   * {@link ExprTreeVisitor} implements an expression visitor for Verilog static checker.
   */
  private final class ExprTreeVisitor extends ExprTreeVisitorDefault {
    private final VerilogNode scope;
    private final boolean implicitDeclaration;

    /**
     * Constructs an expression visitor.
     *
     * @param scope the visitor scope.
     */
    public ExprTreeVisitor(final VerilogNode scope, final boolean implicitDeclaration) {
      InvariantChecks.checkNotNull(scope);

      this.scope = scope;
      this.implicitDeclaration = implicitDeclaration;
    }

    @Override
    public void onOperationBegin(final NodeOperation expr) {

      if (expr.getOperationId() == StandardOperation.FUNCTION) {
        checkFunctionCall(scope, (VerilogExpression) expr.getUserData());
      }
    }

    @Override
    public void onVariable(final NodeVariable variable) {
      checkVariable(scope, variable, implicitDeclaration);
    }
  }

  /** Local declarations to be moved to the global scope. */
  final Map<String, VerilogDeclaration> localDeclarations = new LinkedHashMap<>();

  /** Keeps the block number. */
  private int blockId = 0;

  //================================================================================================
  // Node Visitors
  //================================================================================================

  @Override
  public Result onAssignBegin(final VerilogAssign node) {
    // [IEEE Std 1364-2005]
    //
    // 4.5 Implicit declarations
    //
    // If an identifier appears on the left-hand side of a continuous assignment statement, and
    // that identifier has not been declared previously in the scope where the continuous
    // assignment statement appears or in any scope whose declarations can be directly referenced
    // from the scope where the continuous assignment statement appears (see 12.7), then an
    // implicit scalar net of default net type shall be assumed. See 6.1.2 for a discussion of
    // continuous assignment statements.
    checkReference(node, node.getAssignment(), true /* Implicit declaration */);
    return Result.OK;
  }

  @Override
  public Result onAssignStatementBegin(final VerilogAssignStatement node) {
    checkReference(node, node.getAssignment());
    return Result.OK;
  }


  @Override
  public Result onBlockStatementBegin(final VerilogBlockStatement node) {
    // If the block is unnamed, assign some name.
    if (node.getName() == null) {
      final String blockName = String.format("block.%d", blockId++);
      VerilogLogger.debug("Block name: %s", blockName);

      node.setName(blockName);
    }

    return Result.OK;
  }

  @Override
  public Result onCaseGenerateBegin(final VerilogCaseGenerate node) {
    checkReference(node, node.getExpression());
    return Result.OK;
  }

  @Override
  public Result onCaseStatementBegin(final VerilogCaseStatement node) {
    checkReference(node, node.getExpression());
    return Result.OK;
  }

  @Override
  public Result onCodeBegin(final VerilogCode node) {
    return Result.OK;
  }

  @Override
  public Result onCodeEnd(final VerilogCode node) {
    VerilogLogger.debug("Local declarations: %s", localDeclarations);

    // Move the local declarations to the global scope.
    for (final VerilogDeclaration localDeclaration : localDeclarations.values()) {
      VerilogNode sourceScope = localDeclaration.getParentNode();
      VerilogNode targetScope = sourceScope;

      while (targetScope.getParentNode() != node) {
        targetScope = targetScope.getParentNode();
      }

      VerilogLogger.debug("Move '%s' from %s to %s",
          localDeclaration, localDeclaration.getParentNode(), targetScope);

      // Exclude the declaration from the block.
      localDeclaration.getParentNode().remove(localDeclaration);

      // Rename the declared variable (use the full name except for the module name).
      final String fullName = localDeclaration.getFullName();
      localDeclaration.setName(fullName.substring(fullName.indexOf('.') + 1));

      // Add the declaration to the global scope.
      localDeclaration.setParent(targetScope);
      targetScope.add(localDeclaration);

      // Rename the local variables in the expressions.
      VerilogTransformerVariableRename.transform(sourceScope);
    }

    return Result.OK;
  }

  @Override
  public Result onDeclarationBegin(final VerilogDeclaration node) {
    final VerilogMinTypMax expression = node.getExpression();

    for (final VerilogRange range: node.getDimensions()) {
      checkReference(node, range);
    }

    checkReference(node, node.getElementType());

    if (expression != null) {
      checkReference(node, expression);
    }

    return Result.OK;
  }

  @Override
  public Result onDelayedStatementBegin(final VerilogDelayedStatement node) {
    checkReference(node, node.getEventControl());
    return Result.OK;
  }

  @Override
  public Result onIfGenerateBegin(final VerilogIfGenerate node) {
    checkReference(node, node.getExpression());
    return Result.OK;
  }

  @Override
  public Result onInstantiationBegin(final VerilogInstantiation node) {
    checkInstantiation(node, node);
    return Result.OK;
  }

  @Override
  public Result onIfStatementBegin(final VerilogIfStatement node) {
    checkReference(node, node.getExpression());
    return Result.OK;
  }

  @Override
  public Result onLoopStatementBegin(final VerilogLoopStatement node) {
    final VerilogAssignment initialization = node.getInitialization();
    final VerilogExpression expression = node.getExpression();
    final VerilogAssignment iteration = node.getIteration();

    if (initialization != null) {
      checkReference(node, initialization);
    }
    if (expression != null) {
      checkReference(node, expression);
    }
    if (iteration != null) {
      checkReference(node, iteration);
    }

    return Result.OK;
  }

  @Override
  public Result onModuleBegin(final VerilogModule node) {
    return Result.OK;
  }

  @Override
  public Result onModuleEnd(final VerilogModule node) {
    // The module's ports are checked in onModuleEnd,
    // because of the parameters that can be declared inside the module.
    for (final VerilogPort port : node.getPorts()) {
      checkReference(node, port);
    }

    return Result.OK;
  }

  @Override
  public Result onPortConnectionBegin(final VerilogPortConnection node) {
    // [IEEE Std 1364-2005]
    //
    // 4.5 Implicit declarations
    //
    // If an identifier is used in the terminal list of a primitive instance or a module instance,
    // and that identifier has not been declared previously in the scope where the instantiation
    // appears or in any scope whose declarations can be directly referenced from the scope where
    // the instantiation appears (see 12.7), then an implicit scalar net of default net type shall
    // be assumed.
    final VerilogExpression expression = node.getExpression();

    if (expression != null) {
      checkReference(node.getParentNode(), expression, true /* Implicit declaration */);
    }

    return Result.OK;
  }

  @Override
  public Result onTaskStatementBegin(final VerilogTaskStatement node) {
    checkTaskStatement(node, node);

    for (final VerilogExpression expression : node.getArguments()) {
      checkReference(node, expression);
    }

    return Result.OK;
  }

  @Override
  public Result onWaitStatementBegin(final VerilogWaitStatement node) {
    checkReference(node, node.getExpression());
    return Result.OK;
  }

  //================================================================================================
  // Reference Resolution Methods
  //================================================================================================

  /**
   * Checks the variable.
   *
   * <p>Links the reference with the declaration and refines the variable's data type.</p>
   *
   * @param scope the checking scope.
   * @param node the variable to be checked.
   * @param implicitDeclaration the flag indicating whether implicit declaration is enabled.
   */
  private void checkVariable(
      final VerilogNode scope,
      final NodeVariable node,
      final boolean implicitDeclaration) {
    final VerilogReference reference = ModelUtils.getReference(node);
    InvariantChecks.checkNotNull(reference, String.format("Null reference: %s", node.getName()));

    // Link the reference with the declaration.
    checkReference(scope, reference, implicitDeclaration);
  }

  private void checkRanges(final VerilogNode scope, final List<VerilogRange> ranges) {
    for (final VerilogRange range : ranges) {
      final VerilogExpression left = range.getLeftExpression();
      if (left != null) {
        checkReference(scope, left);
      }
      final VerilogExpression right = range.getRightExpression();
      if (right != null) {
        checkReference(scope, right);
      }
    }
  }

  private VerilogModule getModule(final VerilogNode scope) {
    VerilogNode module = scope;

    while (module != null) {
      if (module.isModule()) {
        return (VerilogModule) module;
      }

      module = module.getParentNode();
    }

    return null;
  }

  /**
   * Checks the reference.
   *
   * <p>Links the reference with the declaration.</p>
   *
   * @param scope the checking scope.
   * @param node the reference to be checked.
   * @param implicitDeclaration the flag indicating whether implicit declaration is enabled.
   */
  private void checkReference(
      final VerilogNode scope,
      final VerilogReference node,
      final boolean implicitDeclaration) {
    InvariantChecks.checkNotNull(scope);
    InvariantChecks.checkNotNull(node);

    VerilogNode entry = findNode(scope, node.getPath());

    // The declaration has not been found.
    if (entry == null) {
      if (implicitDeclaration) {
        final VerilogDeclaration declaration = new VerilogDeclaration(null);
        final VerilogElementType elementType = new VerilogElementType();

        elementType.setType(VerilogElementType.Type.WIRE);
        declaration.setName(node.getPath().toString());
        declaration.setElementType(elementType);

        final VerilogModule module = getModule(scope);
        module.addDeclaration(declaration);
        VerilogLogger.debug("Declaration of '%s' has been added to module '%s'", node, module);

        entry = declaration;
      } else {
        VerilogLogger.error("Declaration of '%s' has not been found", node);
        VerilogTranslator.exit();
      }
    }

    if (entry.isDeclaration()) {
      // The declaration has been found or created implicitly.
      final VerilogDeclaration declaration = (VerilogDeclaration) entry;
      VerilogLogger.debug("Declaration of '%s' has been found in '%s': %s",
          node, declaration.getParentNode().getFullName(), declaration);

      // Store local declarations (to be moved to the global scope).
      if (declaration.getParentNode().isBlockStatement()) {
        localDeclarations.put(declaration.getFullName(), declaration);
      }

      // Link the reference with the declaration.
      node.setDeclaration(declaration);

      // A reference has the following format: name([index]...[index])?([bits])?.
      final List<VerilogRange> indices = node.getIndicesAndBitsSelection();
      checkRanges(scope, indices);

      final List<VerilogRange> dimensions = declaration.getDimensions();
      checkRanges(scope, dimensions);

      if (indices.size() > dimensions.size()) {
        node.setBitsSelection(true);
      }
    } else if (entry.isModule() || entry.isInstantiation()) {
      // Reference to module (e.g., in $dumpvars).
      VerilogLogger.debug("Module/instance reference has been found: %s", node);
    } else {
      VerilogLogger.error("Declaration of '%s' has not been found", node);
      VerilogTranslator.exit();
    }
  }

  /**
   * Checks the reference.
   *
   * <p>Links the reference with the declaration.</p>
   *
   * @param scope the checking scope.
   * @param node the reference to be checked.
   */
  private void checkReference(final VerilogNode scope, final VerilogReference node) {
    checkReference(scope, node, false);
  }

  /**
   * Checks the expression.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the expression to be checked.
   * @param implicitDeclaration the flag indicating whether implicit declaration is enabled.
   */
  private void checkReference(
      final VerilogNode scope,
      final VerilogExpression node,
      final boolean implicitDeclaration) {
    final ExprTreeVisitor visitor = new ExprTreeVisitor(scope, implicitDeclaration);
    final ExprTreeWalker walker = new ExprTreeWalker(visitor);

    walker.visitNode(node.getNode());
  }

  /**
   * Checks the expression.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the expression to be checked.
   */
  private void checkReference(final VerilogNode scope, final VerilogExpression node) {
    checkReference(scope, node, false);
  }

  /**
   * Checks the extended expression.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the expression to be checked.
   */
  private void checkReference(final VerilogNode scope, final VerilogMinTypMax node) {
    final VerilogExpression min = node.getMinExpression();
    final VerilogExpression typ = node.getTypExpression();
    final VerilogExpression max = node.getMaxExpression();

    if (min != null) {
      checkReference(scope, min);
    }
    if (typ != null) {
      checkReference(scope, typ);
    }
    if (max != null) {
      checkReference(scope, max);
    }
  }

  /**
   * Checks the assignment.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the assignment to be checked.
   * @param implicitDeclaration the flag indication whether implicit declaration is enabled.
   */
  private void checkReference(
      final VerilogNode scope,
      final VerilogAssignment node,
      final boolean implicitDeclaration) {
    for (final VerilogReference reference : node.getReferences()) {
      checkReference(scope, reference, implicitDeclaration);
    }

    final VerilogMinTypMax rhs = node.getRhsExpression();

    if (rhs != null) {
      checkReference(scope, node.getRhsExpression());
    }
  }

  /**
   * Checks the assignment.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the assignment to be checked.
   */
  private void checkReference(final VerilogNode scope, final VerilogAssignment node) {
    checkReference(scope, node, false);
  }

  /**
   * Checks the delay.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the delay control to be checked.
   */
  private void checkReference(final VerilogNode scope, final VerilogDelay node) {
    for (final VerilogMinTypMax expression : node.getDelays()) {
      checkReference(scope, expression);
    }
  }

  /**
   * Checks the event control.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the event control to be checked.
   */
  private void checkReference(final VerilogNode scope, final VerilogEventControl node) {
    final VerilogDelay delay = node.getDelay();
    final List<VerilogEvent> events = node.getEvents();
    final VerilogExpression repeat = node.getRepeat();

    if (delay != null) {
      checkReference(scope, delay);
    }
    if (events != null) {
      for (final VerilogEvent event: node.getEvents()) {
        checkReference(scope, event.getExpression());
      }
    }
    if (repeat != null) {
      checkReference(scope, repeat);
    }
  }

  /**
   * Checks the range.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the range to be checked.
   */
  private void checkReference(final VerilogNode scope, final VerilogRange node) {
    checkReference(scope, node.getLeftExpression());
    checkReference(scope, node.getRightExpression());
  }

  /**
   * Checks the element type.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the element type to be checked.
   */
  private void checkReference(final VerilogNode scope, final VerilogElementType node) {
    final VerilogRange range = node.getRange();

    if (range != null) {
      checkReference(scope, range);
    }
  }

  /**
   * Checks the port.
   *
   * <p>Links the references with the declarations.</p>
   *
   * @param scope the checking scope.
   * @param node the port to be checked.
   */
  private void checkReference(final VerilogNode scope, final VerilogPort node) {
    for (final VerilogReference reference : node.getReferences()) {
      checkReference(scope, reference);
    }
  }

  /**
   * Checks the function call.
   *
   * <p>Links the function call with the declaration.</p>
   *
   * @param scope the checking scope.
   * @param node the function call to be checked.
   */
  private void checkFunctionCall(final VerilogNode scope, final VerilogExpression node) {
    final String name = node.getPath().toString();
    final VerilogProcedure entry = (VerilogProcedure) findNode(scope, node.getPath(), Tag.PROCEDURE);

    // The declaration has been found.
    if (entry != null && entry.isFunction()) {
      // Link the function call with the declaration.
      VerilogLogger.debug("Function declaration '%s' has been found: %s", name, entry);
      node.setDeclaration(entry);

      // Stores the declaration in the first operand.
      final NodeOperation operation = (NodeOperation) node.getNode();
      final NodeValue operand = (NodeValue) operation.getOperand(0);

      operand.getData().setUserData(entry);
    } else {
      VerilogLogger.error("Function declaration '%s' has not been found", name);
      VerilogTranslator.exit();
    }
  }

  /**
   * Checks the task statement.
   *
   * <p>Links the task statement with the declaration.</p>
   *
   * @param scope the checking scope.
   * @param node the task statement to be checked.
   */
  private void checkTaskStatement(final VerilogNode scope, final VerilogTaskStatement node) {
    final String name = node.getPath().toString();
    final VerilogProcedure entry = (VerilogProcedure) findNode(scope, node.getPath(), Tag.PROCEDURE);

    // The declaration has been found.
    if (entry != null && entry.isTask()) {
      // Link the task enabling statement with the declaration.
      VerilogLogger.debug("Task declaration '%s' has been found: %s", name, entry);
      node.setDeclaration(entry);
    } else {
      VerilogLogger.error("Task declaration '%s' has not been found", name);
      VerilogTranslator.exit();
    }
  }

  /**
   * Checks the instance.
   *
   * <p>Links the instance with the declaration.</p>
   *
   * @param scope the checking scope.
   * @param node the task statement to be checked.
   */
  private void checkInstantiation(final VerilogNode scope, final VerilogInstantiation node) {
    final String name = node.getModuleName();
    final VerilogNode entry = (VerilogNode) scope.findRecursively(name, VerilogNode.Tag.MODULE);

    VerilogLogger.debug("Instance: %s", node.getName());

    // The declaration has been found.
    if (entry != null) {
      final VerilogModule module = (VerilogModule) entry;

      VerilogLogger.debug("Module: %s", module.getName());

      // Link the instantiation with the module declaration.
      node.setDeclaration(module);

      // If port connections are specified without port names, set the names.
      final Collection<AbstractNode<Tag>> portConnections = node.items(Tag.PORT_CONNECTION);
      final List<VerilogPort> modulePorts = module.getPorts();

      int instancePortIndex = 0;
      for (final AbstractNode<Tag> portConnection : portConnections) {
        final VerilogPortConnection instancePort = (VerilogPortConnection) portConnection;
        VerilogLogger.debug("Port connection: %s", instancePort.getName());

        if (instancePort.getName() == null && instancePortIndex < modulePorts.size()) {
          final VerilogPort modulePort = modulePorts.get(instancePortIndex);
          final VerilogReference modulePortReference = modulePort.getReferences().iterator().next();
          final String modulePortName = modulePortReference.getPath().toString();

          VerilogLogger.debug("Port connection: set %s", modulePortName);
          instancePort.setName(modulePortName);
        }

        instancePortIndex++;
      }
    }
  }

  private VerilogNode findNode(final VerilogNode scope, final VerilogPath path) {
    return findNode(scope, path, null);
  }

  private VerilogNode findNode(final VerilogNode scope, final VerilogPath path, final Tag tag) {
    final List<VerilogPathItem> items = path.getPath();
    InvariantChecks.checkNotEmpty(items);

    VerilogLogger.debug("Finding %s in %s...", path, scope);

    // Iterate over the path items except for the last one.
    VerilogNode current = scope;

    for (int i = 0; i < items.size() - 1; i++) {
      final VerilogPathItem item = items.get(i);
      final String name = item.getName();

      // NOTE: The find-around method searches not only in the current and the upper tables,
      // but also looks up the neighboring scopes.
      VerilogNode node = (VerilogNode) current.findAroundRecursively(name);

      if (node == null) {
        VerilogLogger.error("Node '%s' has not been found (%s)", name, path);
        VerilogTranslator.exit();
      } else if (node.hasScope() || node.isTransparent()) {
        current = node;
      } else if (node.isInstantiation()) {
        final VerilogInstantiation instance = (VerilogInstantiation) node;
        final String module = instance.getModuleName();

        current = (VerilogNode) current.findRecursively(module, VerilogNode.Tag.MODULE);

        if (current == null) {
          VerilogLogger.error("Module '%s' has not been found (%s)", module, path);
          VerilogTranslator.exit();
        }
      } else {
        VerilogLogger.error("Node '%s' is not a scope or an instance (%s)", node, path);
        VerilogTranslator.exit();
      }
    }

    final VerilogPathItem item = items.get(items.size() - 1);
    final String name = item.getName();

    return (tag != null)
        ? (VerilogNode) current.findRecursively(name, tag)
        : (VerilogNode) current.findRecursively(name);
  }
}
