/*
 * Copyright 2013-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.core;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Stack;

/**
 * {@link TreeWalker} implements a walker that traverses an abstract syntax tree.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public class TreeWalker<Tag extends Enum<Tag>> {

  /** {@link Entry} is a traversal stack entry. */
  private class Entry {
    /** The node being traversed. */
    private AbstractNode<Tag> node;
    /** The iterator over the node's children. */
    private Iterator<? extends AbstractNode<Tag>> iterator;

    /**
     * Constructs a traversal stack entry.
     * 
     * @param node the node to be traversed.
     * @param iterator the iterator over the node's children
     */
    public Entry(
        final AbstractNode<Tag> node,
        final Iterator<? extends AbstractNode<Tag>> iterator) {
      InvariantChecks.checkNotNull(node);
      InvariantChecks.checkNotNull(iterator);

      this.node = node;
      this.iterator = iterator;
    }
  }

  /** The root of the tree to be traversed. */
  private final AbstractNode<Tag> root;
  /** The user-defined visitor of tree nodes. */
  private final NodeVisitor<Tag> visitor;

  /** The DFS traversal stack. */
  private Stack<Entry> stack = new Stack<>();

  /**
   * Constructs a tree walker.
   * 
   * @param root the root of the tree to be traversed.
   * @param visitor the user-defined visitor of tree nodes.
   */
  public TreeWalker(final AbstractNode<Tag> root, final NodeVisitor<Tag> visitor) {
    InvariantChecks.checkNotNull(root);
    InvariantChecks.checkNotNull(visitor);

    this.root = root;
    this.visitor = visitor;
  }

  /**
   * Starts DFS traversal.
   */
  public final void start() {
    onBegin(root);

    while (!stack.isEmpty()) {
      final Entry entry = stack.peek();

      if (entry.iterator != null && entry.iterator.hasNext()) {
        onBegin(entry.iterator.next());
      } else {
        onEnd(entry.node);
      }
    }
  }

  /**
   * Starts visiting the node (calls the {@code onBegin} method).
   * 
   * @param node the node to be visited.
   */
  private void onBegin(final AbstractNode<Tag> node) {
    InvariantChecks.checkNotNull(node);

    switch (visitor.onBegin(node)) {
      case OK:
        // NOTE: This hack is to avoid ConcurrentModificationException,
        // when the visitor adds new nodes (e.g., implicit declarations).
        final Collection<AbstractNode<Tag>> children = new ArrayList<>(node.items());
        stack.push(new Entry(node, children.iterator()));
        break;
      case PRUNE:
        stack.push(new Entry(node, /* Do not traverse children */ null));
        break;
      case REPEAT:
        assert false;
        break;
      case ABORT:
        stack.clear();
        break;
    }
  }

  /**
   * Stops visiting the node (calls the {@code onEnd} method).
   * 
   * @param node the node to be visited.
   */
  private void onEnd(final AbstractNode<Tag> node) {
    InvariantChecks.checkNotNull(node);

    switch (visitor.onEnd(node)) {
      case OK:
        stack.pop();
        break;
      case PRUNE:
        assert false;
        break;
      case REPEAT:
        stack.pop();
        onBegin(node);
        break;
      case ABORT:
        stack.clear();
        break;
    }
  }
}
