/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.elaborator;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.VerilogLogger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * {@link VerilogDesign} represents an elaborated design.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogDesign {
  public static final class Builder {
    /** Design name. */
    private String name;

    /** Maps parameter instances to their values. */
    private final Map<String, VerilogParameter> parameters = new LinkedHashMap<>();
    /** Maps new variable instances to their declarations. */
    private final Map<VerilogVariable, VerilogInstantiator> variables = new LinkedHashMap<>();
    /** Contains process instantiations. */
    private final Collection<VerilogInstantiator> processes = new ArrayList<>();

    public void setName(final String name) {
      InvariantChecks.checkNotNull(name);
      this.name = name;
    }

    public Map<String, VerilogParameter> getParameters() {
      return parameters;
    }

    public void addParameter(
        final VerilogParameter.Type type,
        final String name,
        final Data data) {
      InvariantChecks.checkNotNull(type);
      InvariantChecks.checkNotNull(name);
      InvariantChecks.checkNotNull(data);
  
      final VerilogParameter parameter = parameters.get(name);
  
      if (parameter == null || parameter.getType().getPriority() <= type.getPriority()) {
        parameters.put(name, new VerilogParameter(type, name, data));
      }
    }
  
    public void addVariable(
        final VerilogVariable variable,
        final VerilogInstantiator instantiator) {
      InvariantChecks.checkNotNull(variable);
      InvariantChecks.checkNotNull(instantiator);

      variables.put(variable, instantiator);
    }

    public void addProcess(final VerilogInstantiator instantiator) {
      InvariantChecks.checkNotNull(instantiator);
      processes.add(instantiator);
    }

    /**
     * Calculates the variables' bit sizes and modifies the expressions.
     * 
     * @return the elaborated design.
     */
    public VerilogDesign build() {
      final Map<String, VerilogDescriptor> elaboratedVariables = new LinkedHashMap<>();
      final List<VerilogVariable> designVariables = new ArrayList<>(variables.keySet());
      VerilogLogger.debug("Design variables: %s", designVariables);

      // Refine the variables' data types.
      for (final VerilogVariable variable : designVariables) {
        final VerilogInstantiator instantiator = variables.get(variable);
        InvariantChecks.checkNotNull(instantiator);

        // NOTE: To construct correct initialization expression, variable descriptors are used.
        final VerilogDescriptor descriptor =
            instantiator.instantiateDescriptor(variable, elaboratedVariables, parameters);
        InvariantChecks.checkTrue(descriptor != null && descriptor.getDataType() != DataType.UNKNOWN,
            String.format("Cannot refine type for '%s' (%s)", variable, descriptor.getDeclaration()));

        VerilogLogger.debug("DESCRIPTOR: %s %s (%s)",
            variable.getGlobalName(), descriptor, descriptor.getDeclaration().getType());

        variable.getGlobalNode().setData(descriptor.getDataType().valueUninitialized());
        elaboratedVariables.put(variable.getGlobalName(), descriptor);
      }

      VerilogLogger.debug("Elaborated variables: %s", elaboratedVariables);
      return new VerilogDesign(name, parameters, elaboratedVariables, processes);
    }
  }

  /** Design name. */
  private String name;

  /** Maps parameter instances to their values. */
  private final Map<String, VerilogParameter> parameters;
  /** Maps variable instances to their declarations. */
  private final Map<String, VerilogDescriptor> variables;
  /** Contains process instantiations. */
  private final Collection<VerilogInstantiator> instantiators;

  private VerilogDesign(
      final String name,
      final Map<String, VerilogParameter> parameters,
      final Map<String, VerilogDescriptor> variables,
      final Collection<VerilogInstantiator> instantiators) {
    // An empty design's name is null.
    InvariantChecks.checkNotNull(parameters);
    InvariantChecks.checkNotNull(variables);
    InvariantChecks.checkNotNull(instantiators);

    this.name = name;
    this.parameters = Collections.unmodifiableMap(parameters);
    this.variables = Collections.unmodifiableMap(variables);
    this.instantiators = Collections.unmodifiableCollection(instantiators);
  }

  /**
   * Returns the name of the design.
   * 
   * @return the design name.
   */
  public String getName() {
    return name;
  }

  public Map<String, VerilogParameter> getParameters() {
    return parameters;
  }

  public Map<String, VerilogDescriptor> getVariables() {
    return variables;
  }

  /**
   * Returns the iterator over the design processes.
   * 
   * @return the process iterator.
   */
  public Iterable<VerilogProcess> getProcesses() {
    return new Iterable<VerilogProcess>() {
      @Override
      public Iterator<VerilogProcess> iterator() {
        return new Iterator<VerilogProcess>() {
          private final Iterator<VerilogInstantiator> iterator = instantiators.iterator();

          @Override
          public boolean hasNext() {
            return iterator.hasNext();
          }

          @Override
          public VerilogProcess next() {
            // Processes are instantiated on the fly.
            final VerilogInstantiator instantiator = iterator.next();
            // TODO: Implement instance caching.
            return instantiator.instantiateProcess(variables, parameters);
          }
        };
      }
    };
  }
}
