/*
 * Copyright 2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.verilog.parser.elaborator;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link VerilogVariable} stores information about a design variable.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogVariable {
  public enum Type {
    /** Real variable (requires memory). */
    VARIABLE,
    /** Binding (does not require memory). */
    BINDING
  }

  /** Variable type. */
  private final Type type;
  /** Old variable instance (local name). */
  private final NodeVariable oldNode;
  /** New variable instance (global name). */
  private final NodeVariable newNode;

  public VerilogVariable(final Type type, final NodeVariable oldNode, final NodeVariable newNode) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(oldNode);
    InvariantChecks.checkNotNull(newNode);

    this.type = type;
    this.oldNode = oldNode;
    this.newNode = newNode;
  }

  public Type getType() {
    return type;
  }

  public NodeVariable getLocalNode() {
    return oldNode;
  }

  public String getLocalName() {
    return oldNode.getName();
  }

  public NodeVariable getGlobalNode() {
    return newNode;
  }

  public String getGlobalName() {
    return newNode.getName();
  }

  @Override
  public int hashCode() {
    return newNode.hashCode();
  }

  @Override
  public boolean equals(final Object o) {
    if (this == o) {
      return true;
    }

    if (!(o instanceof VerilogVariable)) {
      return false;
    }

    final VerilogVariable r = (VerilogVariable) o;
    return newNode.equals(r.newNode);
  }

  @Override
  public String toString() {
    return newNode.toString();
  }
}
