/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.interpreter;

import ru.ispras.fortress.calculator.CalculatorEngine;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Reducer;
import ru.ispras.fortress.transformer.ValueProvider;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.VerilogLogger;

import java.util.Map;

/**
 * {@link VerilogCalculator} implements expression reduction and evaluation utilities.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogCalculator {

  /**
   * Reduces the expression by folding constant subexpressions.
   * 
   * @param node the expression to be reduced.
   * @return the reduced expression.
   */
  public static Node reduce(final Node node) {
    InvariantChecks.checkNotNull(node);

    VerilogLogger.debug("Reduce: %s", node);

    final CalculatorEngine calculator = VerilogOperations.getCalculatorEngine();
    return Reducer.reduce(calculator, ReduceOptions.NEW_INSTANCE, node);
  }

  /**
   * Evaluates the expression.
   * 
   * @param node the expression to be evaluated.
   * @param valueProvider the variables' values provider.
   * @return the expression value or {@code null}.
   */
  public static Data evaluate(final Node node, final ValueProvider valueProvider) {
    InvariantChecks.checkNotNull(node);

    final Node result;
    final CalculatorEngine calculator = VerilogOperations.getCalculatorEngine();

    if (valueProvider != null) {
      result = Reducer.reduce(calculator, valueProvider, ReduceOptions.NEW_INSTANCE, node);
    } else {
      result = Reducer.reduce(calculator, ReduceOptions.NEW_INSTANCE, node);
    }

    if (result.getKind() != Node.Kind.VALUE) {
      return null;
    }

    final NodeValue value = (NodeValue) result;
    return value.getData();
  }

  /**
   * Evaluates the expression.
   * 
   * @param node the expression to be evaluated.
   * @param context the variables' values.
   * @return the expression value or {@code null}.
   */
  public static Data evaluate(final Node node, final Map<String, Data> context) {
    return evaluate(node, new ValueProvider() {
      @Override
      public Data getVariableValue(final Variable variable) {
        return context.get(variable.getName());
      }
    });
  }

  /**
   * Evaluates the expression.
   * 
   * @param node the expression to be evaluated.
   * @return the expression value or {@code null}.
   */
  public static Data evaluate(final Node node) {
    InvariantChecks.checkNotNull(node);

    final CalculatorEngine calculator = VerilogOperations.getCalculatorEngine();
    final Node result = Reducer.reduce(calculator, ReduceOptions.NEW_INSTANCE, node);

    if (result.getKind() != Node.Kind.VALUE) {
      return null;
    }

    final NodeValue value = (NodeValue) result;
    return value.getData();
  }

  /**
   * Evaluates the integer-value expression.
   * 
   * @param node the expression to be evaluated.
   * @return the expression value or {@code null}.
   */
  public static Integer evaluateInteger(final Node node) {
    final Data data = evaluate(node);
    return toInteger(data);
  }

  /**
   * Evaluates the integer-value expression.
   * 
   * @param node the expression to be evaluated.
   * @param valueProvider the variables' values provider.
   * @return the expression value or {@code null}.
   */
  public static Integer evaluateInteger(final Node node, final ValueProvider valueProvider) {
    final Data data = evaluate(node, valueProvider);
    return toInteger(data);
  }

  /**
   * Evaluates the integer-value expression.
   * 
   * @param node the expression to be evaluated.
   * @param context the variables' values.
   * @return the expression value or {@code null}.
   */
  public static Integer evaluateInteger(final Node node, final Map<String, Data> context) {
    final Data data = evaluate(node, context);
    return toInteger(data);
  }

  private static Integer toInteger(final Data data) {
    if (data == null) {
      return null;
    } else if (data.isType(DataTypeId.LOGIC_BOOLEAN)) {
      return data.getBoolean() ? 1 : 0;
    } else if (data.isType(DataTypeId.LOGIC_INTEGER)) {
      return data.getInteger().intValue();
    } else if (data.isType(DataTypeId.BIT_VECTOR)) {
      return data.getBitVector().intValue();
    }

    return null;
  }

  private VerilogCalculator() {}
}
