/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.interpreter;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.model.VerilogAssignStatement;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.VerilogBlockStatement;
import ru.ispras.verilog.parser.model.VerilogCaseStatement;
import ru.ispras.verilog.parser.model.VerilogCaseStatementItem;
import ru.ispras.verilog.parser.model.VerilogDelayedStatement;
import ru.ispras.verilog.parser.model.VerilogDisableStatement;
import ru.ispras.verilog.parser.model.VerilogIfStatement;
import ru.ispras.verilog.parser.model.VerilogLoopStatement;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogNullStatement;
import ru.ispras.verilog.parser.model.VerilogStatement;
import ru.ispras.verilog.parser.model.VerilogTaskStatement;
import ru.ispras.verilog.parser.model.VerilogTriggerStatement;
import ru.ispras.verilog.parser.model.VerilogWaitStatement;

import java.util.Collection;
import java.util.Map;

/**
 * {@link VerilogInterpreter} implements a simple interpreter of Verilog statements.
 * 
 * <p>
 * The main purpose of {@link VerilogInterpreter} is to evaluate constant expressions that include
 * user-defined functions.
 * </p>
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogInterpreter {

  /**
   * Interprets the given statement in the given context.
   * 
   * @param statement the statement to be interpreted.
   * @param context the values of variables.
   * @return {@code true} iff the statement has been successfully interpreted.
   */
  public static boolean interpret(
      final VerilogStatement statement,
      final Map<String, Data> context) {
    InvariantChecks.checkNotNull(statement);
    InvariantChecks.checkNotNull(context);

    switch (statement.getTag()) {
      case ASSIGN_STATEMENT:
        return interpret((VerilogAssignStatement) statement, context);
      case BLOCK_STATEMENT:
        return interpret((VerilogBlockStatement) statement, context);
      case CASE_STATEMENT:
        return interpret((VerilogCaseStatement) statement, context);
      case DELAYED_STATEMENT:
        return interpret((VerilogDelayedStatement) statement, context);
      case DISABLE_STATEMENT:
        return interpret((VerilogDisableStatement) statement, context);
      case IF_STATEMENT:
        return interpret((VerilogIfStatement) statement, context);
      case LOOP_STATEMENT:
        return interpret((VerilogLoopStatement) statement, context);
      case NULL_STATEMENT:
        return interpret((VerilogNullStatement) statement, context);
      case TASK_STATEMENT:
        return interpret((VerilogTaskStatement) statement, context);
      case TRIGGER_STATEMENT:
        return interpret((VerilogTriggerStatement) statement, context);
      case WAIT_STATEMENT:
        return interpret((VerilogWaitStatement) statement, context);
      default:
        return false;
    }
  }

  private static boolean interpret(
      final VerilogAssignStatement statement,
      final Map<String, Data> context) {
    return interpret(statement.getAssignment(), context);
  }

  private static boolean interpret(
      final VerilogAssignment assignment,
      final Map<String, Data> context) {
    VerilogLogger.debug("Interpret assignment: %s", context);

    final Node lhs = assignment.getLhsNode();
    final Node rhs = assignment.getRhsExpression().getMinExpression().getNode();

    final Data value = VerilogCalculator.evaluate(rhs, context);

    if (lhs.getKind() != Node.Kind.VARIABLE || value == null) {
      return false;
    }

    final NodeVariable variable = (NodeVariable) lhs;
    context.put(variable.getName(), value);

    return true;
  }

  private static boolean interpret(
      final VerilogBlockStatement statement,
      final Map<String, Data> context) {
    VerilogLogger.debug("Interpret block: %s", context);

    final Collection<AbstractNode<VerilogNode.Tag>> items = statement.items();

    for (final AbstractNode<VerilogNode.Tag> item : items) {
      if (!interpret((VerilogStatement) item, context)) {
        return false;
      }
    }

    return true;
  }

  private static boolean interpret(
      final VerilogCaseStatement statement,
      final Map<String, Data> context) {
    VerilogLogger.debug("Interpret case: %s", context);

    final Collection<VerilogCaseStatementItem> items = statement.getCases();

    for (final VerilogCaseStatementItem item : items) {
      final Node node = item.getExpression().getNode();
      final Integer value = VerilogCalculator.evaluateInteger(node, context);

      if (value == null) {
        return false;
      }

      if (value != 0) {
        return interpret(item.getStatement(), context);
      }
    }

    return false;
  }

  private static boolean interpret(
      final VerilogDelayedStatement statement,
      final Map<String, Data> context) {
    // Unsupported statement.
    return false;
  }

  private static boolean interpret(
      final VerilogDisableStatement statement,
      final Map<String, Data> context) {
    // Unsupported statement.
    return false;
  }

  private static boolean interpret(
      final VerilogIfStatement statement,
      final Map<String, Data> context) {
    VerilogLogger.debug("Interpret if-then: %s", context);

    final Node node = statement.getExpression().getNode();
    final Integer value = VerilogCalculator.evaluateInteger(node, context);

    if (value == null) {
      return false;
    }

    return interpret(
        (value != 0 ? statement.getThenStatement() : statement.getElseStatement()), context);
  }

  private static boolean interpret(
      final VerilogLoopStatement statement,
      final Map<String, Data> context) {
    switch (statement.getType()) {
      case FOR:
        return interpretFor(statement, context);
      case FOREVER:
        return interpretForever(statement, context);
      case REPEAT:
        return interpretRepeat(statement, context);
      case WHILE:
        return interpretWhile(statement, context);
      default:
        return false;
    }
  }

  private static boolean interpretFor(
      final VerilogLoopStatement statement,
      final Map<String, Data> context) {
    VerilogLogger.debug("Interpret for: %s", context);

    if (!interpret(statement.getInitialization(), context)) {
      return false;
    }

    final Node node = statement.getExpression().getNode();

    while (true) {
      final Integer value = VerilogCalculator.evaluateInteger(node, context);

      if (value == null) {
        return false;
      }

      if (value == 0) {
        break;
      }

      if (!interpret(statement.getStatement(), context)) {
        return false;
      }

      if (!interpret(statement.getIteration(), context)) {
        return false;
      }
    }

    return true;
  }

  private static boolean interpretForever(
      final VerilogLoopStatement statement,
      final Map<String, Data> context) {
    // Unsupported statement.
    return false;
  }

  private static boolean interpretRepeat(
      final VerilogLoopStatement statement,
      final Map<String, Data> context) {
    VerilogLogger.debug("Interpret repeat: %s", context);

    final Node node = statement.getExpression().getNode();
    final Integer value = VerilogCalculator.evaluateInteger(node, context);

    if (value == null) {
      return false;
    }

    for (int i = 0; i < value; i++) {
      if (!interpret(statement.getStatement(), context)) {
        return false;
      }
    }

    return true;
  }

  private static boolean interpretWhile(
      final VerilogLoopStatement statement,
      final Map<String, Data> context) {
    VerilogLogger.debug("Interpret while: %s", context);

    final Node node = statement.getExpression().getNode();

    while (true) {
      final Integer value = VerilogCalculator.evaluateInteger(node, context);

      if (value == null) {
        return false;
      }

      if (value == 0) {
        break;
      }

      if (!interpret(statement.getStatement(), context)) {
        return false;
      }
    }

    return true;
  }

  private static boolean interpret(
      final VerilogNullStatement statement,
      final Map<String, Data> context) {
    // Do nothing.
    return true;
  }

  private static boolean interpret(
      final VerilogTaskStatement statement,
      final Map<String, Data> context) {
    // Unsupported statement.
    return false;
  }

  private static boolean interpret(
      final VerilogTriggerStatement statement,
      final Map<String, Data> context) {
    // Unsupported statement.
    return false;
  }

  private static boolean interpret(
      final VerilogWaitStatement statement,
      final Map<String, Data> context) {
    // Unsupported statement.
    return false;
  }

  private VerilogInterpreter() {}
}
