/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogCaseGenerateItem} represents generate case items.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogCaseGenerateItem extends VerilogNode {
  public static final Tag TAG = Tag.CASE_GENERATE_ITEM;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(Tag.BLOCK_GENERATE);

  /** Case expressions (to be matched with the case selection expression). */
  private List<VerilogExpression> expressions = new ArrayList<>();

  /**
   * Creates a generate case item.
   * 
   * @param parent the parent node.
   */
  public VerilogCaseGenerateItem(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.PSEUDO_ELEMENT, parent);
  }

  /**
   * Creates a copy of the generate case item.
   * 
   * @param other the generate case item to be copied.
   * @param parent the parent node.
   */
  public VerilogCaseGenerateItem(final VerilogCaseGenerateItem other, final VerilogNode parent) {
    super(other, parent);

    for (final VerilogExpression expression : other.expressions) {
      this.expressions.add(expression.clone());
    }
  }

  /**
   * Returns the expressions of the case.
   * 
   * @return the list of expressions.
   */
  public List<VerilogExpression> getExpressions() {
    return expressions;
  }

  /**
   * Adds the expression to the case.
   * 
   * @param expression the expression to be added.
   */
  public void addExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    expressions.add(expression);
  }

  /**
   * Returns the generate block of the case.
   * 
   * @return the generate block.
   */
  public VerilogBlockGenerate getGenerate() {
    return (VerilogBlockGenerate) items(TAGS_CHILDREN).iterator().next();
  }

  /**
   * Sets the generate block of the case.
   * 
   * @param generate the generate block to be set.
   */
  public void setGenerate(final VerilogBlockGenerate generate) {
    add(generate);
  }

  /**
   * Returns the condition associated with the case.
   * 
   * @return the case condition.
   */
  public VerilogExpression getExpression() {
    // Create a result.
    final VerilogExpression expression = new VerilogExpression();
    // Get the parent node.
    final VerilogCaseGenerate parent = (VerilogCaseGenerate)getParent();

    final VerilogExpression a = parent.getExpression();

    // A non-default case.
    for (final VerilogExpression b : expressions) {
      // In contrast to the case statement, EQ (==) is used instead of EQCASE (===). 
      final NodeOperation equality = new NodeOperation(StandardOperation.EQ,
          a.getNode(), b.getNode());

      if (expression.isNull()) {
        // case(a) b ... => (a === b).
        expression.setNode(equality);
      } else {
        // case(a) b, c ... => (a === b) || (a === c).
        expression.setNode(
            new NodeOperation(StandardOperation.OR, expression.getNode(), equality));
      }
    }

    // A default case.
    if (expression.isNull()) {
      // Iterate over all cases.
      for (final VerilogCaseGenerateItem item : parent.getCases()) {
        // Except the default one.
        if (item == this) {
          continue;
        }

        // Construct the disjunction of all non-default conditions.
        final VerilogExpression b = item.getExpression();

        if (expression.isNull()) {
          // case(a) b ... => non-default = (a === b).
          expression.setNode(b.getNode());
        } else {
          // case(a) b, c ... => non-default = ((a === b) || (a === c)).
          expression.setNode(
              new NodeOperation(StandardOperation.OR, expression.getNode(), b.getNode()));
        }
      }

      // Negate the non-default condition.
      if (expression.isNull()) {
        expression.setNode(NodeValue.newBoolean(true));
      } else {
        expression.setNode(new NodeOperation(StandardOperation.NOT, expression.getNode()));
      }
    }

    return expression;
  }

  @Override
  public VerilogCaseGenerateItem clone() {
    return new VerilogCaseGenerateItem(this, getParentNode());
  }
}
