/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogCaseStatement} represents case statements.
 * 
 * <p>There are the following process types:</p>
 * <ul>
 * <li>{@code case} (ordinary case statement);</li>
 * <li>{@code casex} (X-don't-care case statement);</li>
 * <li>{@code casez} (Z-don't-care case statement).</li>
 * </ul>
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogCaseStatement extends VerilogStatement {
  public static final Tag TAG = Tag.CASE_STATEMENT;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(Tag.CASE_STATEMENT_ITEM);

  /** {@link Type} contains the case statement types. */
  public enum Type {
    /** Ordinary case statement: {@code case ... endcase}. */
    CASE,
    /** X-don't-care case statement: {@code casex ... endcase}. */
    CASEX,
    /** Z-don't-care case statement: {@code casez ... endcase}. */
    CASEZ
  }

  /** Type of the case statement (by default, {@code CASE}). */
  private Type type;
  /** Case selection expression (to be matched with available cases). */
  private VerilogExpression expression;

  /**
   * Creates a case statement.
   * 
   * @param parent the parent node.
   */
  public VerilogCaseStatement(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Creates a copy of the case statement.
   * 
   * @param other the case statement to be copied.
   * @param parent the parent node.
   */
  public VerilogCaseStatement(final VerilogCaseStatement other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;
    this.expression = other.expression != null ? other.expression.clone() : null;
  }

  /**
   * Returns the case statement type.
   * 
   * @return the case statement type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the case statement type.
   * 
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the statement is {@code CASE}.
   * 
   * @return {@code true} iff the statement is {@code CASE}.
   */
  public boolean isCase() {
    return type == Type.CASE;
  }

  /**
   * Checks whether the statement is {@code CASEX}.
   * 
   * @return {@code true} iff the statement is {@code CASEX}.
   */
  public boolean isCaseX() {
    return type == Type.CASEX;
  }

  /**
   * Checks whether the statement is {@code CASEZ}.
   * 
   * @return {@code true} iff the statement is {@code CASEZ}.
   */
  public boolean isCaseZ() {
    return type == Type.CASEZ;
  }

  /**
   * Sets the type of the statement to {@code CASE}.
   */
  public void setCase() {
    type = Type.CASE;
  }

  /**
   * Sets the type of the statement to {@code CASEX}.
   */
  public void setCaseX() {
    type = Type.CASEX;
  }

  /**
   * Sets the type of the statement to {@code CASEZ}.
   */
  public void setCaseZ() {
    type = Type.CASEZ;
  }

  /**
   * Returns the case selection expression.
   * 
   * @return the expression.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the case selection expression.
   * 
   * @param expression the expression to be set.
   */
  public void setExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  /**
   * Returns the cases.
   * 
   * @return the list of cases.
   */
  public List<VerilogCaseStatementItem> getCases() {
    final Collection<AbstractNode<VerilogNode.Tag>> cases = items(TAGS_CHILDREN);
    return new ArrayList<>(Arrays.asList(cases.toArray(new VerilogCaseStatementItem[]{})));
  }

  /**
   * Adds the case to the case selection construct.
   * 
   * @param item the case to be added.
   */
  public void addCase(final VerilogCaseStatementItem item) {
    add(item);
  }

  @Override
  public VerilogCaseStatement clone() {
    return new VerilogCaseStatement(this, getParentNode());
  }
}
