/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogEventControl;

import java.util.EnumSet;

/**
 * {@link VerilogDelayedStatement} represents delayed statements.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogDelayedStatement extends VerilogStatement {
  public static final Tag TAG = Tag.DELAYED_STATEMENT;
  public static final EnumSet<Tag> TAGS_CHILDREN = VerilogStatement.TAGS;

  /** Delay. */
  private VerilogEventControl control;

  /**
   * Creates a delayed statement.
   * 
   * @param parent the parent node.
   */
  public VerilogDelayedStatement(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Creates a delayed statement.
   * 
   * @param other the delayed statement to be copied.
   * @param parent the parent node.
   */
  public VerilogDelayedStatement(final VerilogDelayedStatement other, final VerilogNode parent) {
    super(other, parent);

    this.control = other.control != null ? other.control.clone() : null;
  }

  /**
   * Returns the delay of the delayed statement.
   * 
   * @return the delay.
   */
  public VerilogEventControl getEventControl() {
    return control;
  }

  /**
   * Sets the delay of the delayed statement.
   * 
   * @param control the delay.
   */
  public void setEventControl(final VerilogEventControl control) {
    InvariantChecks.checkNotNull(control);
    this.control = control;
  }

  /**
   * Returns the statement of the delayed statement.
   * 
   * @return the statement.
   */
  public VerilogStatement getStatement() {
    return (VerilogStatement) items(TAGS_CHILDREN).iterator().next();
  }

  /**
   * Sets the statement of the delayed statement.
   * 
   * @param statement the statement.
   */
  public void setStatement(final VerilogStatement statement) {
    add(statement);
  }

  @Override
  public VerilogDelayedStatement clone() {
    return new VerilogDelayedStatement(this, getParentNode());
  }
}
