/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.EnumSet;

/**
 * {@link VerilogIfStatement} represents if-then-else statements.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogIfStatement extends VerilogStatement {
  public static final Tag TAG = Tag.IF_STATEMENT;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(Tag.IF_STATEMENT_BRANCH);

  /** Condition. */
  private VerilogExpression expression;

  /**
   * Creates an if-then-else statement.
   * 
   * @param parent the parent node.
   */
  public VerilogIfStatement(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Creates a copy of the if-then-else statement.
   * 
   * @param other the if-then-else statement to be copied.
   * @param parent the parent node.
   */
  public VerilogIfStatement(final VerilogIfStatement other, final VerilogNode parent) {
    super(other, parent);

    this.expression = other.expression != null ? other.expression.clone() : null;
  }

  /**
   * Returns the condition.
   * 
   * @return the condition.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the condition.
   * 
   * @param expression the condition.
   */
  public void setExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  /**
   * Returns the statement related to the branch {@code then}.
   * 
   * @return the {@code then}'s statement.
   */
  public VerilogStatement getThenStatement() {
    for (final AbstractNode<VerilogNode.Tag> item : items(TAGS_CHILDREN)) {
      final VerilogIfStatementBranch branch = (VerilogIfStatementBranch) item;

      if (branch.isThen()) {
        return branch.getStatement();
      }
    }

    return null;
  }

  /**
   * Sets the statement related to the branch {@code then}.
   * 
   * @param thenStatement the statement to be set.
   */
  public void setThenStatement(final VerilogStatement thenStatement) {
    InvariantChecks.checkNotNull(thenStatement);

    final VerilogIfStatementBranch branch =
        new VerilogIfStatementBranch(VerilogIfStatementBranch.Type.THEN, thenStatement, this);

    add(branch);
  }

  /**
   * Returns the statement related to the branch {@code else}.
   * 
   * @return the {@code else}'s statement.
   */
  public VerilogStatement getElseStatement() {
    for (final AbstractNode<VerilogNode.Tag> item : items(TAGS_CHILDREN)) {
      final VerilogIfStatementBranch branch = (VerilogIfStatementBranch) item;

      if (branch.isElse()) {
        return branch.getStatement();
      }
    }

    return null;
  }

  /**
   * Sets the statement related to the branch {@code else}.
   * 
   * @param elseStatement the statement to be set.
   */
  public void setElseStatement(final VerilogStatement elseStatement) {
    InvariantChecks.checkNotNull(elseStatement);

    final VerilogIfStatementBranch branch =
        new VerilogIfStatementBranch(VerilogIfStatementBranch.Type.ELSE, elseStatement, this);

    add(branch);
  }

  @Override
  public VerilogIfStatement clone() {
    return new VerilogIfStatement(this, getParentNode());
  }
}
