/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.EnumSet;

/**
 * {@link VerilogIfStatementBranch} represents if-then-else statement branches (then and else).
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogIfStatementBranch extends VerilogNode {
  public static final Tag TAG = Tag.IF_STATEMENT_BRANCH;
  public static final EnumSet<Tag> TAGS_CHILDREN = VerilogStatement.TAGS;

  /** {@link Type} contains the branch types. */
  public enum Type {
    /** Branch {@code then} (the condition is satisfied). */
    THEN,
    /** Branch {@code else} (the condition is not satisfied). */
    ELSE
  }

  /** Type of the branch (by default, {@code THEN}). */
  private Type type = Type.THEN;

  /**
   * Creates an if-then-else statement branch.
   * 
   * @param parent the parent node.
   */
  public VerilogIfStatementBranch(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.ELEMENT, parent);
  }

  /**
   * Create an if-then-else statement branch.
   * 
   * @param type the branch type.
   * @param statement the branch body. 
   * @param parent the parent node.
   */
  public VerilogIfStatementBranch(
      final Type type,
      final VerilogStatement statement,
      final VerilogNode parent) {
    this(parent);

    setType(type);
    setStatement(statement);
  }

  public VerilogIfStatementBranch(final VerilogIfStatementBranch other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;
  }

  /**
   * Returns the branch type ({@code THEN} or {@code ELSE}).
   * 
   * @return the branch type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the branch type.
   * 
   * @param type the branch type.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the branch is {@code THEN}.
   * 
   * @return {@code true} iff the branch is {@code THEN}.
   */
  public boolean isThen() {
    return type == Type.THEN;
  }

  /**
   * Checks whether the branch is {@code ELSE}.
   * 
   * @return {@code true} iff the branch is {@code ELSE}.
   */
  public boolean isElse() {
    return type == Type.ELSE;
  }

  /**
   * Sets the branch type to {@code THEN}.
   */
  public void setThen() {
    type = Type.THEN;
  }

  /**
   * Sets the branch type to {@code ELSE}.
   */
  public void setElse() {
    type = Type.ELSE;
  }

  /**
   * Returns the branch body.
   * 
   * @return the branch body.
   */
  public VerilogStatement getStatement() {
    return (VerilogStatement) items(TAGS_CHILDREN).iterator().next();
  }

  /**
   * Sets the branch body.
   * 
   * @param statement the branch body.
   */
  public void setStatement(final VerilogStatement statement) {
    add(statement);
  }

  /**
   * Returns the condition associated with the branch.
   * 
   * @return the branch condition.
   */
  public VerilogExpression getExpression() {
    // Create a result.
    final VerilogExpression expression = new VerilogExpression();
    // Get the parent node.
    final VerilogIfStatement parent = (VerilogIfStatement)getParent();

    final VerilogExpression a = parent.getExpression();

    if (isThen()) {
      // if (a) then ... -> a.
      expression.setNode(a.getNode());
    } else {
      // if (a) else ...-> !a.
      expression.setNode(new NodeOperation(StandardOperation.NOT, a.getNode()));
    }

    return expression;
  }

  @Override
  public VerilogIfStatementBranch clone() {
    return new VerilogIfStatementBranch(this, getParentNode());
  }
}
