/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogPath;
import ru.ispras.verilog.parser.model.basis.VerilogPathItem;
import ru.ispras.verilog.parser.model.basis.VerilogReference;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogModule} represents the abstract syntax of the module declaration.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogModule extends VerilogNode {
  public static final Tag TAG = Tag.MODULE;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(
      // Module declarations (variables and parameters).
      Tag.DECLARATION,
      // Processes, instances and procedures.
      Tag.ACTIVITY,
      Tag.ASSIGN,
      Tag.INSTANTIATION,
      Tag.PROCEDURE,
      Tag.TABLE,
      // Parameters' overrides.
      Tag.DEFINE_PARAMETER,
      // Specify constructs.
      Tag.SPECIFY,
      // Generate constructs.
      Tag.GENERATE,
      Tag.LOOP_GENERATE,
      Tag.IF_GENERATE,
      Tag.CASE_GENERATE
  );

  /** {@link Type} contains the module types. */
  public enum Type {
    /** Module. */
    MODULE,
    /** Macro module */
    MACROMODULE,
    /** User-defined primitive (UDP). */
    USER_PRIMITIVE,
    /** Gate primitive (AND, OR, etc.). */
    GATE_PRIMITIVE
  }

  /** Type of the module. */
  private Type type = Type.MODULE;
  /** Signature of the module (list of ports). */
  private List<VerilogPort> ports = new ArrayList<>();

  /**
   * Creates a module.
   * 
   * @param parent the parent node.
   */
  public VerilogModule(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.CLASS, parent);
  }

  /**
   * Creates a copy of the module.
   * 
   * @param other the module to be copied.
   * @param parent the parent node.
   */
  public VerilogModule(final VerilogModule other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;

    for (final VerilogPort port : other.ports) {
      this.ports.add(port.clone());
    }
  }

  /**
   * Returns the type of the module.
   * 
   * @return the type of the module.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the type of the module.
   * 
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the module is {@code MODULE}.
   * 
   * @return {@code true} iff the module is {@code MODULE}.
   */
  public boolean isModule() {
    return type == Type.MODULE;
  }

  /**
   * Checks whether the module is {@code MACROMODULE}.
   * 
   * @return {@code true} iff the module is {@code MACROMODULE}.
   */
  public boolean isMacromodule() {
    return type == Type.MACROMODULE;
  }

  /**
   * Checks whether the module is {@code USER_PRIMITIVE}.
   * 
   * @return {@code true} iff the module is {@code USER_PRIMITIVE}.
   */
  public boolean isUserPrimitive() {
    return type == Type.USER_PRIMITIVE;
  }

  /**
   * Checks whether the module is {@code GATE_PRIMITIVE}.
   * 
   * @return {@code true} iff the module is {@code GATE_PRIMITIVE}.
   */
  public boolean isGatePrimitive() {
    return type == Type.GATE_PRIMITIVE;
  }

  /**
   * Sets the type of the module to {@code MODULE}.
   */
  public void setModule() {
    type = Type.MODULE;
  }

  /**
   * Sets the type of the module to {@code MACROMODULE}.
   */
  public void setMacromodule() {
    type = Type.MACROMODULE;
  }

  /**
   * Sets the type of the module to {@code USER_PRIMITIVE}.
   */
  public void setUserPrimitive() {
    type = Type.USER_PRIMITIVE;
  }

  /**
   * Sets the type of the module to {@code GATE_PRIMITIVE}.
   */
  public void setGatePrimitive() {
    type = Type.GATE_PRIMITIVE;
  }

  /**
   * Returns the signature of the module (i.e., the list of its ports).
   * 
   * @return the signature of the module.
   */
  public List<VerilogPort> getPorts() {
    return ports;
  }

  /**
   * Adds the port to the signature of the module.
   * 
   * @param port the port to be added.
   */
  public void addPort(final VerilogPort port) {
    InvariantChecks.checkNotNull(port);
    ports.add(port);
  }

  /**
   * Adds the declaration to the module.
   * 
   * @param declaration the declaration to be added.
   */
  public void addDeclaration(final VerilogDeclaration declaration) {
    add(declaration);

    // Add the declared ports into the signature.
    if (declaration.isPort()) {
      final VerilogPort port = new VerilogPort(this);

      final VerilogReference reference = new VerilogReference();
      final VerilogPath path = new VerilogPath();
      final VerilogPathItem item = new VerilogPathItem(declaration.getName());

      path.addItem(item);
      reference.setPath(path);
      port.addReference(reference);

      addPort(port);
    }
  }

  @Override
  public VerilogModule clone() {
    return new VerilogModule(this, getParentNode());
  }
}
