/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogReference;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogShowCancelled} represents show-cancelled constructs.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogShowCancelled extends VerilogNode {
  public static final Tag TAG = Tag.SHOW_CANCELLED;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** {@link VerilogShowCancelled} contains the show-cancelled construct types. */
  public enum Type {
    /** Show: {@code showcancelled x}. */
    SHOW,
    /** No show: {@code noshowcancelled x}. */
    NO_SHOW
  }

  /** Type of the show-cancelled construct. (by default, {@code SHOW}) */
  private Type type = Type.SHOW;
  /** References under consideration. */
  private List<VerilogReference> references = new ArrayList<>();

  /**
   * Creates a show-cancelled construct.
   * 
   * @param parent the parent node.
   */
  public VerilogShowCancelled(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.OBJECT, parent);
  }

  /**
   * Creates a copy of the show-cancelled construct.
   * 
   * @param other the show-cancelled construct to be copied. 
   * @param parent the parent node.
   */
  public VerilogShowCancelled(final VerilogShowCancelled other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;

    for (final VerilogReference reference : other.references) {
      this.references.add(reference.clone());
    }
  }

  /**
   * Returns the show-cancelled construct type ({@code SHOW} or {@code NO_SHOW}).
   * 
   * @return the construct type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the show-cancelled construct type ({@code SHOW} or {@code NO_SHOW}).
   * 
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the construct is {@code SHOW}.
   * 
   * @return {@code true} iff the construct is {@code SHOW}.
   */
  public boolean isShow() {
    return type == Type.SHOW;
  }

  /**
   * Checks whether the construct is {@code NO_SHOW}.
   * 
   * @return {@code true} iff the construct is {@code NO_SHOW}.
   */
  public boolean isNoShow() {
    return type == Type.NO_SHOW;
  }

  /**
   * Sets the construct type to {@code SHOW}.
   */
  public void setShow() {
    type = Type.SHOW;
  }

  /**
   * Sets the construct type to {@code NO_SHOW}.
   */
  public void setNoShow() {
    type = Type.NO_SHOW;
  }

  /**
   * Returns the references of the show-cancelled construct.
   * 
   * @return the list of references.
   */
  public List<VerilogReference> getReferences() {
    return references;
  }

  /**
   * Adds the reference to the show-cancelled construct.
   * 
   * @param reference the reference to be added.
   */
  public void addReference(final VerilogReference reference) {
    InvariantChecks.checkNotNull(reference);
    this.references.add(reference);
  }

  @Override
  public VerilogShowCancelled clone() {
    return new VerilogShowCancelled(this, getParentNode());
  }
}
