/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogPath;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogTriggerStatement} represents trigger statements.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogTriggerStatement extends VerilogStatement {
  public static final Tag TAG = Tag.TRIGGER_STATEMENT;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** Path of the trigger statement. */
  private VerilogPath path;
  /** Arguments of the trigger statement. */
  private List<VerilogExpression> arguments = new ArrayList<>();

  /**
   * Creates a trigger statement.
   * 
   * @param parent the parent node.
   */
  public VerilogTriggerStatement(final VerilogNode parent) {
    super(TAG, parent);
  }

  /**
   * Creates a copy of the trigger statement.
   * 
   * @param other the trigger statement to be coppied.
   * @param parent the parent node.
   */
  public VerilogTriggerStatement(final VerilogTriggerStatement other, final VerilogNode parent) {
    super(other, parent);

    this.path = other.path != null ? other.path.clone() : null;

    for (final VerilogExpression expression : other.arguments) {
      arguments.add(expression.clone());
    }
  }

  /**
   * Returns the path to the trigger statement.
   * 
   * @return the task path.
   */
  public VerilogPath getPath() {
    return path;
  }

  /**
   * Sets the path of the trigger statement.
   * 
   * @param path the path to be set.
   */
  public void setPath(final VerilogPath path) {
    InvariantChecks.checkNotNull(path);
    this.path = path;
  }

  /**
   * Returns the arguments of the trigger statement.
   * 
   * @return the arguments.
   */
  public List<VerilogExpression> getArguments() {
    return arguments;
  }

  /**
   * Adds the argument to the trigger statement.
   * 
   * @param argument the argument to be added.
   */
  public void addArgument(final VerilogExpression argument) {
    InvariantChecks.checkNotNull(argument);
    arguments.add(argument);
  }

  @Override
  public VerilogTriggerStatement clone() {
    return new VerilogTriggerStatement(this, getParentNode());
  }
}
