/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link VerilogPathItem} represents path items.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogPathItem {
  /** Instance name. */
  private String name;
  /** Instance index. */
  private VerilogExpression expression;

  /**
   * Creates a path item with no name.
   */
  public VerilogPathItem() {}

  /**
   * Creates a path item with the given instance name.
   * 
   * @param name the instance name.
   */
  public VerilogPathItem(final String name) {
    InvariantChecks.checkNotNull(name);
    this.name = name;
  }

  /**
   * Creates a path item with the given name and index.
   * 
   * @param name the instance name.
   * @param expression the instance index.
   */
  public VerilogPathItem(final String name, final VerilogExpression expression) {
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(expression);

    this.name = name;
    this.expression = expression;
  }

  /**
   * Creates a copy of the path item.
   * 
   * @param other the path item to be copied.
   */
  public VerilogPathItem(final VerilogPathItem other) {
    InvariantChecks.checkNotNull(other);

    this.name = other.name != null ? other.name : null;
    this.expression = other.expression != null ? other.expression.clone() : null;
  }

  /**
   * Returns the instance name of the path item.
   * 
   * @return the instance name.
   */
  public String getName() {
    return name;
  }

  /**
   * Sets the instance name of the path item.
   * 
   * @param name the instance name to be set.
   */
  public void setName(final String name) {
    InvariantChecks.checkNotNull(name);
    this.name = name;
  }

  /**
   * Returns the instance index of the path item.
   * 
   * @return the instance index.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the instance index of the path item.
   * 
   * @param expression the instance index to be set.
   */
  public void setExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  @Override
  public String toString() {
    return expression != null
        ? String.format("%s[%s]", name, expression)
        : name;
  }

  @Override
  public VerilogPathItem clone() {
    return new VerilogPathItem(this);
  }
}
