/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.List;

/**
 * {@link VerilogRepeatEvents} represents repeat-event constructs.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogRepeatEvents {
  /** Repeat count. */
  private VerilogExpression expression;
  /** Events to be repeated. */
  private List<VerilogEvent> events;

  /**
   * Creates a repeat-event construct.
   */
  public VerilogRepeatEvents() {}

  /**
   * Creates a copy of the repeat-event construct.
   * 
   * @param other the repeat-event construct to be copied.
   */
  public VerilogRepeatEvents(final VerilogRepeatEvents other) {
    InvariantChecks.checkNotNull(other);

    this.expression = other.expression.clone();

    this.events = new ArrayList<>(other.events.size());
    for (final VerilogEvent event : other.events) {
      this.events.add(event.clone());
    }
  }

  /**
   * Returns the repeat count expression.
   * 
   * @return the repeat count.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets a repeat count expression.
   * 
   * @param expression the expression to be set.
   */
  public void setExpression(final VerilogExpression expression) {
    this.expression = expression;
  }

  /**
   * Returns the list of events to be repeated.
   * 
   * @return the list of events.
   */
  public List<VerilogEvent> getEvents() {
    return events;
  }

  /**
   * Sets a list of events to be repeated.
   * 
   * @param events the list to be set.
   */
  public void setEvents(final List<VerilogEvent> events) {
    this.events = events;
  }

  @Override
  public String toString() {
    return String.format("(%s) %s", expression, events);
  }

  @Override
  public VerilogRepeatEvents clone() {
    return new VerilogRepeatEvents(this);
  }
}
