/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link VerilogStrength} represents signal strengths.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogStrength {

  /**
   * {@link Type} contains the strengths.
   * 
   *  <p>Strengths are used to resolve which value should appear on a net or a gate output.</p>
   */
  public enum Type {
    /** Supply strength (the strongest one). */
    SUPPLY,
    /** Strong strength. */
    STRONG,
    /** Pull strength. */
    PULL,
    /** Large strength. */
    LARGE,
    /** Weak strength. */
    WEAK,
    /** Medium strength. */
    MEDIUM,
    /** Small strength. */
    SMALL,
    /** High-Z strength (the weakest one). */
    HIGHZ
  }

  /** Drive strength for 0. */
  private Type strength0;
  /** Drive strength for 1. */
  private Type strength1;
  /** Charge strength. */
  private Type strength;

  /**
   * Creates a signal strength.
   */
  public VerilogStrength() {}

  /**
   * Creates a copy of the signal strength.
   * 
   * @param other the signal strength to be copied.
   */
  public VerilogStrength(final VerilogStrength other) {
    InvariantChecks.checkNotNull(other);

    this.strength0 = other.strength0;
    this.strength1 = other.strength1;
    this.strength = other.strength;
  }

  /**
   * Returns the drive strength for 0.
   * 
   * @return the drive strength for 0.
   */
  public Type getStrength0() {
    return strength0;
  }

  /**
   * Sets the drive strength for 0.
   * 
   * @param type the drive strength for 0.
   */
  public void setStrength0(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.strength0 = type;
  }

  /**
   * Checks whether the drive strength for 0 is {@code SUPPLY}.
   * 
   * @return {@code true} iff the drive strength for 0 is {@code SUPPLY}.
   */
  public boolean isSupply0() {
    return strength0 == Type.SUPPLY;
  }

  /**
   * Checks whether the drive strength for 0 is {@code STRONG}.
   * 
   * @return {@code true} iff the drive strength for 0 is {@code STRONG}.
   */
  public boolean isStrong0() {
    return strength0 == Type.STRONG;
  }

  /**
   * Checks whether the drive strength for 0 is {@code PULL}.
   * 
   * @return {@code true} iff the drive strength for 0 is {@code PULL}.
   */
  public boolean isPull0() {
    return strength0 == Type.PULL;
  }

  /**
   * Checks whether the drive strength for 0 is {@code LARGE}.
   * 
   * @return {@code true} iff the drive strength for 0 is {@code LARGE}.
   */
  public boolean isLarge0() {
    return strength0 == Type.LARGE;
  }

  /**
   * Checks whether the drive strength for 0 is {@code WEAK}.
   * 
   * @return {@code true} iff the drive strength for 0 is {@code WEAK}.
   */
  public boolean isWeak0() {
    return strength0 == Type.WEAK;
  }

  /**
   * Checks whether the drive strength for 0 is {@code MEDIUM}.
   * 
   * @return {@code true} iff the drive strength for 0 is {@code MEDIUM}.
   */
  public boolean isMedium0() {
    return strength0 == Type.MEDIUM;
  }

  /**
   * Checks whether the drive strength for 0 is {@code SMALL}.
   * 
   * @return {@code true} iff the drive strength for 0 is {@code SMALL}.
   */
  public boolean isSmall0() {
    return strength0 == Type.SMALL;
  }

  /**
   * Checks whether the drive strength for 0 is {@code HIGHZ}.
   * 
   * @return {@code true} iff the drive strength for 0 is {@code HIGHZ}.
   */
  public boolean isHighZ0() {
    return strength0 == Type.HIGHZ;
  }

  /**
   * Sets the drive strength for 0 to {@code SUPLLY}.
   */
  public void setSupply0() {
    strength0 = Type.SUPPLY;
  }

  /**
   * Sets the drive strength for 0 to {@code STRONG}.
   */
  public void setStrong0() {
    strength0 = Type.STRONG;
  }

  /**
   * Sets the drive strength for 0 to {@code PULL}.
   */
  public void setPull0() {
    strength0 = Type.PULL;
  }

  /**
   * Sets the drive strength for 0 to {@code LARGE}.
   */
  public void setLarge0() {
    strength0 = Type.LARGE;
  }

  /**
   * Sets the drive strength for 0 to {@code WEAK}.
   */
  public void setWeak0() {
    strength0 = Type.WEAK;
  }

  /**
   * Sets the drive strength for 0 to {@code MEDIUM}.
   */
  public void setMedium0() {
    strength0 = Type.MEDIUM;
  }

  /**
   * Sets the drive strength for 0 to {@code SMALL}.
   */
  public void setSmall0() {
    strength0 = Type.SMALL;
  }

  /**
   * Sets the drive strength for 0 to {@code HIGHZ}.
   */
  public void setHighZ0() {
    strength0 = Type.HIGHZ;
  }

  /**
   * Returns the drive strength for 1.
   * 
   * @return the drive strength for 1.
   */
  public Type getStrength1() {
    return strength1;
  }

  /**
   * Sets the drive strength for 1.
   * 
   * @param type the drive strength for 1.
   */
  public void setStrength1(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.strength1 = type;
  }

  /**
   * Checks whether the drive strength for 1 is {@code SUPPLY}.
   * 
   * @return {@code true} iff the drive strength for 1 is {@code SUPPLY}.
   */
  public boolean isSupply1() {
    return strength1 == Type.SUPPLY;
  }

  /**
   * Checks whether the drive strength for 1 is {@code STRONG}.
   * 
   * @return {@code true} iff the drive strength for 1 is {@code STRONG}.
   */
  public boolean isStrong1() {
    return strength1 == Type.STRONG;
  }

  /**
   * Checks whether the drive strength for 1 is {@code PULL}.
   * 
   * @return {@code true} iff the drive strength for 1 is {@code PULL}.
   */
  public boolean isPull1() {
    return strength1 == Type.PULL;
  }

  /**
   * Checks whether the drive strength for 1 is {@code LARGE}.
   * 
   * @return {@code true} iff the drive strength for 1 is {@code LARGE}.
   */
  public boolean isLarge1() {
    return strength1 == Type.LARGE;
  }

  /**
   * Checks whether the drive strength for 1 is {@code WEAK}.
   * 
   * @return {@code true} iff the drive strength for 1 is {@code WEAK}.
   */
  public boolean isWeak1() {
    return strength1 == Type.WEAK;
  }

  /**
   * Checks whether the drive strength for 1 is {@code MEDIUM}.
   * 
   * @return {@code true} iff the drive strength for 1 is {@code MEDIUM}.
   */
  public boolean isMedium1() {
    return strength1 == Type.MEDIUM;
  }

  /**
   * Checks whether the drive strength for 1 is {@code SMALL}.
   * 
   * @return {@code true} iff the drive strength for 1 is {@code SMALL}.
   */
  public boolean isSmall1() {
    return strength1 == Type.SMALL;
  }

  /**
   * Checks whether the drive strength for 1 is {@code HIGHZ}.
   * 
   * @return {@code true} iff the drive strength for 1 is {@code HIGHZ}.
   */
  public boolean isHighZ1() {
    return strength1 == Type.HIGHZ;
  }

  /**
   * Sets the drive strength for 1 to {@code SUPPLY}.
   */
  public void setSupply1() {
    strength1 = Type.SUPPLY;
  }

  /**
   * Sets the drive strength for 1 to {@code STRONG}.
   */
  public void setStrong1() {
    strength1 = Type.STRONG;
  }

  /**
   * Sets the drive strength for 1 to {@code PULL}.
   */
  public void setPull1() {
    strength1 = Type.PULL;
  }

  /**
   * Sets the drive strength for 1 to {@code LARGE}.
   */
  public void setLarge1() {
    strength1 = Type.LARGE;
  }

  /**
   * Sets the drive strength for 1 to {@code WEAK}.
   */
  public void setWeak1() {
    strength1 = Type.WEAK;
  }

  /**
   * Sets the drive strength for 1 to {@code MEDIUM}.
   */
  public void setMedium1() {
    strength1 = Type.MEDIUM;
  }

  /**
   * Sets the drive strength for 1 to {@code SMALL}.
   */
  public void setSmall1() {
    strength1 = Type.SMALL;
  }

  /**
   * Sets the drive strength for 1 to {@code HIGHZ}.
   */
  public void setHighZ1() {
    strength1 = Type.HIGHZ;
  }

  /**
   * Returns the charge strength.
   * 
   * @return the charge strength.
   */
  public Type getStrength() {
    return strength;
  }

  /**
   * Sets the charge strength.
   * 
   * @param type the charge strength.
   */
  public void setStrength(final Type type) {
    this.strength = type;
  }

  /**
   * Checks whether the charge strength is {@code SMALL}.
   * 
   * @return {@code true} iff the charge strength is {@code SMALL}.
   */
  public boolean isSmall() {
    return strength == Type.SMALL;
  }

  /**
   * Checks whether the charge strength is {@code MEDIUM}.
   * 
   * @return {@code true} iff the charge strength is {@code MEDIUM}.
   */
  public boolean isMedium() {
    return strength == Type.MEDIUM;
  }

  /**
   * Checks whether the charge strength is {@code LARGE}.
   * 
   * @return {@code true} iff the charge strength is {@code LARGE}.
   */
  public boolean isLarge() {
    return strength == Type.LARGE;
  }

  /**
   * Sets the charge strength to {@code SMALL}.
   */
  public void setSmall() {
    strength = Type.SMALL;
  }

  /**
   * Sets the charge strength to {@code MEDIUM}.
   */
  public void setMedium() {
    strength = Type.MEDIUM;
  }

  /**
   * Sets the charge strength to {@code LARGE}.
   */
  public void setLarge() {
    strength = Type.LARGE;
  }

  @Override
  public String toString() {
    return String.format("%s:%s:%s", strength0, strength1, strength);
  }

  @Override
  public VerilogStrength clone() {
    return new VerilogStrength(this);
  }
}
